# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from typing import Any


class IpCoreFile:
    """
    Class for handling an IP core file.
    """

    def __init__(self, path: Path, **variables: Any) -> None:
        """
        Arguments:
            path: Path to the TCL script that creates the IP.
                Should typically end in .tcl.
            variables: These name/value variable pairs will be set in TCL before
                the IP core ``.tcl`` file is sourced. This makes it possible to parameterize the IP
                core creation.

                .. note::
                    This is a "kwargs" style argument. You can pass any number of named arguments.
        """
        self.path = path
        self.variables = variables

    @property
    def name(self) -> str:
        """
        A shorthand name for this IP core.
        """
        file_name_no_suffix = self.path.stem
        if " " in file_name_no_suffix:
            raise ValueError(f"File name may not contain spaces: {self.path}")
        return file_name_no_suffix

    def __str__(self) -> str:
        return f"{self.__class__.__name__}:{self.path}:{self.variables}"



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path

# First party libraries
from tsfpga.constraint import Constraint
from tsfpga.examples.example_env import get_hdl_modules, get_tsfpga_example_modules
from tsfpga.examples.vivado.project import TsfpgaExampleVivadoProject
from tsfpga.module import BaseModule

THIS_FILE = Path(__file__)


class Module(BaseModule):
    def get_build_projects(self):
        projects = []

        modules = get_hdl_modules() + get_tsfpga_example_modules()
        part = "xc7z020clg400-1"

        tcl_dir = self.path / "tcl"
        pinning = Constraint(tcl_dir / "artyz7_pinning.tcl")
        block_design = tcl_dir / "block_design.tcl"

        projects.append(
            TsfpgaExampleVivadoProject(
                name="artyz7",
                modules=modules,
                part=part,
                tcl_sources=[block_design],
                constraints=[pinning],
                defined_at=THIS_FILE,
            )
        )

        projects.append(
            TsfpgaExampleVivadoProject(
                name="artyz7_explore",
                top="artyz7_top",
                modules=modules,
                part=part,
                tcl_sources=[block_design],
                constraints=[pinning],
                impl_explore=True,
                defined_at=THIS_FILE,
            )
        )

        projects.append(
            TsfpgaExampleVivadoProject(
                name="artyz7_verilog",
                top="artyz7_top_verilog",
                modules=modules,
                part=part,
                constraints=[pinning],
                defined_at=THIS_FILE,
            )
        )

        projects.append(
            TsfpgaExampleVivadoProject(
                name="artyz7_systemverilog",
                top="artyz7_top_systemverilog",
                modules=modules,
                part=part,
                constraints=[pinning],
                defined_at=THIS_FILE,
            )
        )

        return projects



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import importlib.util
import os
import subprocess
from os.path import commonpath, relpath
from pathlib import Path
from platform import system
from shutil import rmtree
from typing import Any, Optional

# First party libraries
from tsfpga import DEFAULT_FILE_ENCODING


def create_file(file: Path, contents: Optional[str] = None) -> Path:
    """
    Create the ``file`` and any parent directories that do not exist.
    File will be empty unless ``contents`` is specified.

    Return:
        The path to the file that was created (i.e. the original ``file`` argument).
    """
    # Create directory unless it already exists. Do not delete anything if it does exist.
    create_directory(directory=file.parent, empty=False)

    contents = "" if contents is None else contents
    with open(file, "w", encoding=DEFAULT_FILE_ENCODING) as file_handle:
        file_handle.write(contents)

    return file


def read_file(file: Path) -> str:
    """
    Read and return the file contents.
    """
    with open(file, encoding=DEFAULT_FILE_ENCODING) as file_handle:
        return file_handle.read()


def read_last_lines_of_file(file: Path, num_lines: int) -> str:
    """
    Read a number of lines from the end of a file, without buffering the whole file.
    Similar to unix ``tail`` command.

    Arguments:
        file: The file that shall be read.
        num_lines: The number of lines to read.

    Return:
        The last lines of the file.
    """
    result_lines: list[str] = []
    blocks_to_read = 0

    with open(file, encoding=DEFAULT_FILE_ENCODING) as file_handle:
        while len(result_lines) < num_lines:
            # Since we do not know the line lengths, there is some guessing involved. Keep reading
            # larger and larger blocks until we have all the lines that are requested.
            blocks_to_read += 1

            try:
                # Read a block from the end
                file_handle.seek(-blocks_to_read * 4096, os.SEEK_END)
            except IOError:
                # Tried to read more data than what is available. Read whatever we have and return
                # to user.
                file_handle.seek(0)
                result_lines = file_handle.readlines()
                break

            result_lines = file_handle.readlines()

    result = "".join(result_lines[-num_lines:])
    return result


def delete(path: Path, wait_until_deleted: bool = False) -> Path:
    """
    Delete a file or directory from the filesystem.

    Arguments:
        path: The file/directory to be deleted.
        wait_until_deleted: When set to ``True``, the function will poll the filesystem
            after initiating the deletion, and not return until the path is in fact deleted.
            Is needed on some filesystems/mounts in a situation where we delete a path and
            then directly want to write to it afterwards.

    Return:
        The path that was deleted (i.e. the original ``path`` argument).
    """
    if path.exists():
        if path.is_dir():
            rmtree(path)
        else:
            path.unlink()

    if wait_until_deleted:
        while path.exists():
            pass

    return path


def create_directory(directory: Path, empty: bool = True) -> Path:
    """
    Create a directory.

    Arguments:
        directory: Path to the directory.
        empty: If true and the directory already exists, all existing files/folders in it will
            be deleted.
            If false, the directory will be left as-is if it already exists, or will be newly
            created if it does not.

    Return:
        The path that was created (i.e. the original ``directory`` argument).
    """
    if empty:
        # Delete directory, and anything inside it, before creating it below.
        delete(directory)

    elif directory.exists():
        if directory.is_dir():
            return directory

        raise FileExistsError(f"Requested directory path already exists as a file: {directory}")

    directory.mkdir(parents=True)
    return directory


def file_is_in_directory(file_path: Path, directories: list[Path]) -> bool:
    """
    Check if the file is in any of the directories.

    Arguments:
        file_path: The file to be checked.
        directories: The directories to be controlled.

    Return:
        True if there is a common path.
    """
    for directory in directories:
        if commonpath([str(file_path), str(directory)]) == str(directory):
            return True

    return False


def path_relative_to(path: Path, other: Path) -> Path:
    """
    Return a relative path from ``other`` to ``path``.
    This function works even if ``path`` is not inside a ``other`` folder.

    Note Path.relative_to() does not support the use case where e.g. readme.md should get
    relative path "../readme.md". Hence we have to use os.path.
    """
    return Path(relpath(str(path), str(other)))


def run_command(
    cmd: list[str],
    cwd: Optional[Path] = None,
    env: Optional[dict[str, str]] = None,
    capture_output: bool = False,
) -> subprocess.CompletedProcess[str]:
    """
    Will raise ``CalledProcessError`` if the command fails.

    Arguments:
        cmd: The command to run.
        cwd: The working directory where the command shall be executed.
        env: Environment variables to set.
        capture_output: Enable capturing of STDOUT and STDERR.

    Return:
        Returns the subprocess completed process object, which contains useful information.
        If ``capture_output`` is set, the ``stdout`` and ``stderr`` members of this object can
        be inspected.
    """
    if not isinstance(cmd, list):
        raise ValueError("Must be called with a list, not a string")

    return subprocess.run(
        args=cmd,
        cwd=cwd,
        env=env,
        check=True,
        encoding=DEFAULT_FILE_ENCODING,
        capture_output=capture_output,
    )


def load_python_module(file: Path) -> Any:
    """
    Load the specified Python module.
    Note that in Python nomenclature, a module is a source code file.

    On the returned object, you can call functions or instantiate classes that are in the module.
    """
    python_module_name = file.stem

    spec = importlib.util.spec_from_file_location(name=python_module_name, location=file)
    if spec is None or spec.loader is None:
        raise RuntimeError(f"Could not load the Python module: {file}")

    module = importlib.util.module_from_spec(spec=spec)
    spec.loader.exec_module(module=module)

    return module


def system_is_windows() -> bool:
    """
    Return True if the script is being executed on a computer running the Windows operating system.
    """
    return system() == "Windows"



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# First party libraries
from tsfpga.vivado.build_result import BuildResult


def test_size_summary():
    build_result = BuildResult(name="apa")

    build_result.synthesis_size = dict(LUT=3, FFs=4)
    expected = """\
Size of apa after synthesis:
{
  "LUT": 3,
  "FFs": 4
}"""
    assert build_result.size_summary() == expected

    build_result.implementation_size = dict(LUT=8, FFs=9)
    expected = """\
Size of apa after implementation:
{
  "LUT": 8,
  "FFs": 9
}"""
    assert build_result.size_summary() == expected


def test_report():
    build_result = BuildResult(name="apa")

    build_result.synthesis_size = dict(LUT=3, FFs=4)
    expected = """\
Size of apa after synthesis:
{
  "LUT": 3,
  "FFs": 4
}"""
    assert build_result.report() == expected

    # Add a logic level distribution report
    build_result.logic_level_distribution = """\
+-----------------+-------------+-----+----+---+----+
| End Point Clock | Requirement |  0  |  1 | 2 |  3 |
+-----------------+-------------+-----+----+---+----+
| clk_fpga_0      | 2.000ns     | 491 | 12 | 1 | 11 |
+-----------------+-------------+-----+----+---+----+\
"""
    expected = """\
Size of apa after synthesis:
{
  "LUT": 3,
  "FFs": 4
}
Logic level distribution:
+-----------------+-------------+-----+----+---+----+
| End Point Clock | Requirement |  0  |  1 | 2 |  3 |
+-----------------+-------------+-----+----+---+----+
| clk_fpga_0      | 2.000ns     | 491 | 12 | 1 | 11 |
+-----------------+-------------+-----+----+---+----+\
"""
    assert build_result.report() == expected
    assert build_result.maximum_logic_level == 3


def test_maximum_logic_level_should_be_none_if_no_logic_level_distribution_is_set():
    build_result = BuildResult(name="apa")
    build_result.synthesis_size = dict(LUT=3, FFs=4)

    # maximum_logic_level is calculated based on the logic_level_distribution, which is not set
    # in this test case
    assert build_result.logic_level_distribution is None
    assert build_result.maximum_logic_level is None
    assert "level" not in build_result.report()



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

# Local folder libraries
from .system_utils import create_file, path_relative_to

if TYPE_CHECKING:
    # Local folder libraries
    from .module_list import ModuleList
    from .vivado.simlib_common import VivadoSimlibCommon


def create_ghdl_ls_configuration(
    output_path: Path,
    modules: "ModuleList",
    vunit_proj: Any,
    simlib: Optional["VivadoSimlibCommon"] = None,
) -> None:
    """
    Create a configuration file (hdl-prj.json) for the vhdl-lsp VHDL Language Server
    (https://github.com/ghdl/ghdl-language-server).

    Can be used with modules and an "empty" VUnit project, or with a complete VUnit
    project with all user files added.

    Execution of this function takes roughly 12 ms for a large project (62 modules and a
    VUnit project).

    Arguments:
        output_path: Output folder.
        modules: Source files from these modules will be included.
        vunit_proj: A VUnit project.
        simlib: Source from this Vivado simlib project will be added.
    """

    def get_relative_path(path: Path) -> Path:
        return path_relative_to(path=path, other=output_path)

    data = dict(options=dict(ghdl_analysis=[]), files=[])

    def add_compiled_library(path: Path) -> None:
        relative_path = get_relative_path(path)
        data["options"]["ghdl_analysis"].append(f"-P{relative_path}")  # type: ignore[index]

    data["options"]["ghdl_analysis"] += [  # type: ignore[index]
        "--std=08",
    ]

    # pylint: disable=protected-access
    compiled_vunit_libraries_path = Path(vunit_proj._output_path) / "ghdl" / "libraries"
    for compiled_library_path in compiled_vunit_libraries_path.glob("*"):
        add_compiled_library(compiled_library_path)

    if simlib is not None:
        for library_name in simlib.library_names:
            add_compiled_library(simlib.output_path / library_name)

    # The same file might be present in both module file list as well as VUnit project.
    # However the opposite might also be true in many cases.
    # E.g. files that depend on IP cores that are not included in the simulation project.
    # For this reason, add all files to a set first to avoid duplicates.
    files = set()

    if modules is not None:
        for module in modules:
            for hdl_file in module.get_simulation_files(include_ip_cores=False):
                files.add(hdl_file.path)

    for source_file in vunit_proj.get_compile_order():
        files.add(Path(source_file.name).resolve())

    for file_path in files:
        data["files"].append(  # type: ignore[attr-defined]
            dict(file=str(get_relative_path(file_path)), language="vhdl")
        )

    create_file(output_path / "hdl-prj.json", json.dumps(data))



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import shutil
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

# First party libraries
from tsfpga import TSFPGA_TCL
from tsfpga.build_step_tcl_hook import BuildStepTclHook
from tsfpga.constraint import Constraint
from tsfpga.system_utils import create_file, read_file

# Local folder libraries
from .build_result import BuildResult
from .common import run_vivado_gui, run_vivado_tcl
from .hierarchical_utilization_parser import HierarchicalUtilizationParser
from .logic_level_distribution_parser import LogicLevelDistributionParser
from .tcl import VivadoTcl

if TYPE_CHECKING:
    # First party libraries
    from tsfpga.module_list import ModuleList

    # Local folder libraries
    from .build_result_checker import MaximumLogicLevel, SizeChecker


class VivadoProject:
    """
    Used for handling a Xilinx Vivado HDL project
    """

    # pylint: disable=too-many-arguments,too-many-instance-attributes
    def __init__(
        self,
        name: str,
        modules: "ModuleList",
        part: str,
        top: Optional[str] = None,
        generics: Optional[dict[str, Any]] = None,
        constraints: Optional[list["Constraint"]] = None,
        tcl_sources: Optional[list[Path]] = None,
        build_step_hooks: Optional[list["BuildStepTclHook"]] = None,
        vivado_path: Optional[Path] = None,
        default_run_index: int = 1,
        impl_explore: bool = False,
        defined_at: Optional[Path] = None,
        **other_arguments: Any,
    ):  # pylint: disable=too-many-locals
        """
        Class constructor. Performs a shallow copy of the mutable arguments, so that the user
        can e.g. append items to their list after creating an object.

        Arguments:
            name: Project name.
            modules: Modules that shall be included in the project.
            part: Part identification.
            top: Name of top level entity.
                If left out, the top level name will be inferred from the ``name``.
            generics: A dict with generics values (name: value). Use this parameter
                for "static" generics that do not change between multiple builds of this
                project. These will be set in the project when it is created.

                Compare to the build-time generic argument in :meth:`build`.

                The generic value shall be of type

                * :class:`bool` (suitable for VHDL type ``boolean`` and ``std_logic``),
                * :class:`int` (suitable for VHDL type ``integer``, ``natural``, etc.),
                * :class:`float` (suitable for VHDL type ``real``),
                * :class:`.BitVectorGenericValue` (suitable for VHDL type ``std_logic_vector``,
                  ``unsigned``, etc.), or
                * :class:`.StringGenericValue` (suitable for VHDL type ``string``).
            constraints: Constraints that will be applied to the project.
            tcl_sources: A list of TCL files. Use for e.g. block design, pinning, settings, etc.
            build_step_hooks: Build step hooks that will be applied to the project.
            vivado_path: A path to the Vivado executable.
                If omitted, the default location from the system PATH will be used.
            default_run_index: Default run index (synth_X and impl_X) that is set in the
                project.
                Can also use the argument to :meth:`build() <VivadoProject.build>` to
                specify at build-time.
            defined_at: Optional path to the file where you defined this project.
                To get a useful ``build_fpga.py --list`` message. Is useful when you have many
                projects set up.
            other_arguments: Optional further arguments. Will not be used by tsfpga, but will
                instead be passed on to

                * :func:`BaseModule.get_synthesis_files()
                  <tsfpga.module.BaseModule.get_synthesis_files>`
                * :func:`BaseModule.get_ip_core_files()
                  <tsfpga.module.BaseModule.get_ip_core_files>`
                * :func:`BaseModule.get_scoped_constraints()
                  <tsfpga.module.BaseModule.get_scoped_constraints>`
                * :func:`VivadoProject.pre_create`
                * :func:`BaseModule.pre_build() <tsfpga.module.BaseModule.pre_build>`
                * :func:`VivadoProject.pre_build`
                * :func:`VivadoProject.post_build`

                along with further arguments supplied at build-time to :meth:`.create` and
                :meth:`.build`.

                .. note::
                    This is a "kwargs" style argument. You can pass any number of named arguments.
        """
        self.name = name
        self.modules = modules.copy()
        self.part = part
        self.static_generics = {} if generics is None else generics.copy()
        self.constraints = [] if constraints is None else constraints.copy()
        self.tcl_sources = [] if tcl_sources is None else tcl_sources.copy()
        self.build_step_hooks = [] if build_step_hooks is None else build_step_hooks.copy()
        self._vivado_path = vivado_path
        self.default_run_index = default_run_index
        self.impl_explore = impl_explore
        self.defined_at = defined_at
        self.other_arguments = None if other_arguments is None else other_arguments.copy()

        # Will be set by subclass when applicable
        self.is_netlist_build = False
        self.analyze_synthesis_timing = True
        self.report_logic_level_distribution = False
        self.ip_cores_only = False

        self.top = name + "_top" if top is None else top

        self.tcl = VivadoTcl(name=self.name)

        for constraint in self.constraints:
            if not isinstance(constraint, Constraint):
                raise TypeError(f'Got bad type for "constraints" element: {constraint}')

        for tcl_source in self.tcl_sources:
            if not isinstance(tcl_source, Path):
                raise TypeError(f'Got bad type for "tcl_sources" element: {tcl_source}')

        for build_step_hook in self.build_step_hooks:
            if not isinstance(build_step_hook, BuildStepTclHook):
                raise TypeError(f'Got bad type for "build_step_hooks" element: {build_step_hook}')

    def project_file(self, project_path: Path) -> Path:
        """
        Arguments:
            project_path: A path containing a Vivado project.

        Return:
            The project file of this project, in the given folder
        """
        return project_path / (self.name + ".xpr")

    def _setup_tcl_sources(self) -> None:
        tsfpga_tcl_sources = [
            TSFPGA_TCL / "vivado_default_run.tcl",
            TSFPGA_TCL / "vivado_fast_run.tcl",
            TSFPGA_TCL / "vivado_messages.tcl",
        ]

        if self.impl_explore:
            tsfpga_tcl_sources.append(TSFPGA_TCL / "vivado_strategies.tcl")

        # Add tsfpga TCL sources first. The user might want to change something in the tsfpga
        # settings. Conversely, tsfpga should not modify something that the user has set up.
        self.tcl_sources = tsfpga_tcl_sources + self.tcl_sources

    def _setup_build_step_hooks(self) -> None:
        # Check that no ERROR messages have been sent by Vivado. After synthesis as well as
        # after implementation.
        self.build_step_hooks.append(
            BuildStepTclHook(
                TSFPGA_TCL / "check_no_error_messages.tcl", "STEPS.SYNTH_DESIGN.TCL.POST"
            )
        )
        self.build_step_hooks.append(
            BuildStepTclHook(
                TSFPGA_TCL / "check_no_error_messages.tcl", "STEPS.WRITE_BITSTREAM.TCL.PRE"
            )
        )

        # Check the implemented timing and resource utilization via TCL build hooks.
        # This is different than for synthesis, where it is embedded in the build script.
        # This is due to Vivado limitations related to post-synthesis hooks.
        # Specifically, the report_utilization figures do not include IP cores when it is run in
        # a post-synthesis hook.
        self.build_step_hooks.append(
            BuildStepTclHook(TSFPGA_TCL / "report_utilization.tcl", "STEPS.WRITE_BITSTREAM.TCL.PRE")
        )
        self.build_step_hooks.append(
            BuildStepTclHook(TSFPGA_TCL / "check_timing.tcl", "STEPS.WRITE_BITSTREAM.TCL.PRE")
        )

        if not self.analyze_synthesis_timing:
            # In this special case however, the synthesized design is never opened, and
            # report_utilization is not run by the build_vivado_project.tcl.
            # So in order to get a utilization report anyway we add it as a hook.
            # This mode is exclusively used by netlist builds, which very rarely include IP cores,
            # so it is acceptable that the utilization report might be erroneous with regards to
            # IP cores.
            self.build_step_hooks.append(
                BuildStepTclHook(
                    TSFPGA_TCL / "report_utilization.tcl", "STEPS.SYNTH_DESIGN.TCL.POST"
                )
            )

        if self.report_logic_level_distribution:
            # Used by netlist builds
            self.build_step_hooks.append(
                BuildStepTclHook(
                    TSFPGA_TCL / "report_logic_level_distribution.tcl",
                    "STEPS.SYNTH_DESIGN.TCL.POST",
                )
            )

    def _create_tcl(
        self, project_path: Path, ip_cache_path: Optional[Path], all_arguments: dict[str, Any]
    ) -> Path:
        """
        Make a TCL file that creates a Vivado project
        """
        if project_path.exists():
            raise ValueError(f"Folder already exists: {project_path}")
        project_path.mkdir(parents=True)

        create_vivado_project_tcl = project_path / "create_vivado_project.tcl"
        tcl = self.tcl.create(
            project_folder=project_path,
            modules=self.modules,
            part=self.part,
            top=self.top,
            run_index=self.default_run_index,
            generics=self.static_generics,
            constraints=self.constraints,
            tcl_sources=self.tcl_sources,
            build_step_hooks=self.build_step_hooks,
            ip_cache_path=ip_cache_path,
            disable_io_buffers=self.is_netlist_build,
            ip_cores_only=self.ip_cores_only,
            other_arguments=all_arguments,
        )
        create_file(create_vivado_project_tcl, tcl)

        return create_vivado_project_tcl

    def create(
        self,
        project_path: Path,
        ip_cache_path: Optional[Path] = None,
        **other_arguments: Any,
    ) -> bool:
        """
        Create a Vivado project

        Arguments:
            project_path: Path where the project shall be placed.
            ip_cache_path: Path to a folder where the Vivado IP cache can be
                placed. If omitted, the Vivado IP cache mechanism will not be enabled.
            other_arguments: Optional further arguments. Will not be used by tsfpga, but will
                instead be sent to

                * :func:`BaseModule.get_synthesis_files()
                  <tsfpga.module.BaseModule.get_synthesis_files>`
                * :func:`BaseModule.get_ip_core_files()
                  <tsfpga.module.BaseModule.get_ip_core_files>`
                * :func:`BaseModule.get_scoped_constraints()
                  <tsfpga.module.BaseModule.get_scoped_constraints>`
                * :func:`VivadoProject.pre_create`

                along with further ``other_arguments`` supplied to :meth:`.__init__`.

                .. note::
                    This is a "kwargs" style argument. You can pass any number of named arguments.
        Return:
            True if everything went well.
        """
        print(f"Creating Vivado project in {project_path}")
        self._setup_tcl_sources()
        self._setup_build_step_hooks()

        # The pre-create hook might have side effects. E.g. change some register constants.
        # So we make a deep copy of the module list before the hook is called.
        # Note that the modules are copied before the pre-build hooks as well,
        # since we do not know if we might be performing a create-only or
        # build-only operation. The copy does not take any significant time, so this is not
        # an issue.
        self.modules = deepcopy(self.modules)

        # Send all available arguments that are reasonable to use in pre-create and module getter
        # functions. Prefer run-time values over the static.
        all_arguments = copy_and_combine_dicts(self.other_arguments, other_arguments)
        all_arguments.update(
            generics=self.static_generics,
            part=self.part,
        )

        if not self.pre_create(
            project_path=project_path, ip_cache_path=ip_cache_path, **all_arguments
        ):
            print("ERROR: Project pre-create hook returned False. Failing the build.")
            return False

        create_vivado_project_tcl = self._create_tcl(
            project_path=project_path, ip_cache_path=ip_cache_path, all_arguments=all_arguments
        )
        return run_vivado_tcl(self._vivado_path, create_vivado_project_tcl)

    def pre_create(self, **kwargs: Any) -> bool:  # pylint: disable=unused-argument
        """
        Override this function in a subclass if you wish to do something useful with it.
        Will be called from :meth:`.create` right before the call to Vivado.

        An example use case for this function is when TCL source scripts for the Vivado project
        have to be auto generated. This could e.g. be scripts that set IP repo paths based on the
        Vivado system PATH.

        .. Note::
            This default method does nothing. Shall be overridden by project that utilize
            this mechanism.

        Arguments:
            kwargs: Will have all the :meth:`.create` parameters in it, as well as everything in
                the ``other_arguments`` argument to :func:`VivadoProject.__init__`.

        Return:
            True if everything went well.
        """
        return True

    def _build_tcl(
        self,
        project_path: Path,
        output_path: Path,
        num_threads: int,
        run_index: int,
        all_generics: dict[str, Any],
        synth_only: bool,
        from_impl: bool,
        impl_explore: bool,
    ) -> Path:
        """
        Make a TCL file that builds a Vivado project
        """
        project_file = self.project_file(project_path)
        if not project_file.exists():
            raise ValueError(
                f"Project file does not exist in the specified location: {project_file}"
            )

        build_vivado_project_tcl = project_path / "build_vivado_project.tcl"
        tcl = self.tcl.build(
            project_file=project_file,
            output_path=output_path,
            num_threads=num_threads,
            run_index=run_index,
            generics=all_generics,
            synth_only=synth_only,
            from_impl=from_impl,
            analyze_synthesis_timing=self.analyze_synthesis_timing,
            impl_explore=impl_explore,
        )
        create_file(build_vivado_project_tcl, tcl)

        return build_vivado_project_tcl

    def pre_build(self, **kwargs: Any) -> bool:  # pylint: disable=unused-argument
        """
        Override this function in a subclass if you wish to do something useful with it.
        Will be called from :meth:`.build` right before the call to Vivado.

        Arguments:
            kwargs: Will have all the :meth:`.build` parameters in it. Including additional
                parameters from the user.

        Return:
            True if everything went well.
        """
        return True

    def post_build(self, **kwargs: Any) -> bool:  # pylint: disable=unused-argument
        """
        Override this function in a subclass if you wish to do something useful with it.
        Will be called from :meth:`.build` right after the call to Vivado.

        An example use case for this function is to encrypt the bit file, or generate any other
        material that shall be included in FPGA release artifacts.

        .. Note::
            This default method does nothing. Shall be overridden by project that utilize
            this mechanism.

        Arguments:
            kwargs: Will have all the :meth:`.build` parameters in it. Including additional
                parameters from the user. Will also include ``build_result`` with
                implemented/synthesized size, which can be used for asserting the expected resource
                utilization.

        Return:
            True if everything went well.
        """
        return True

    def build(  # pylint: disable=too-many-locals,too-many-branches
        self,
        project_path: Path,
        output_path: Optional[Path] = None,
        run_index: Optional[int] = None,
        generics: Optional[dict[str, Any]] = None,
        synth_only: bool = False,
        from_impl: bool = False,
        num_threads: int = 12,
        **pre_and_post_build_parameters: Any,
    ) -> BuildResult:
        """
        Build a Vivado project

        Arguments:
            project_path: A path containing a Vivado project.
            output_path: Results (bit file, ...) will be placed here.
            run_index: Select Vivado run (synth_X and impl_X) to build with.
            generics: A dict with generics values (`dict(name: value)`). Use for run-time
                generics, i.e. values that can change between each build of this project.

                Compare to the create-time generics argument in :meth:`.__init__`.

                The generic value types follow the same rules as for :meth:`.__init__`.
            synth_only: Run synthesis and then stop.
            from_impl: Run the ``impl`` steps and onward on an existing synthesized design.
            num_threads: Number of parallel threads to use during run.
            pre_and_post_build_parameters: Optional further arguments. Will not be used by tsfpga,
                but will instead be sent to

                * :func:`BaseModule.pre_build() <tsfpga.module.BaseModule.pre_build>`
                * :func:`VivadoProject.pre_build`
                * :func:`VivadoProject.post_build`

                along with further ``other_arguments`` supplied to :meth:`.__init__`.

                .. note::
                    This is a "kwargs" style argument. You can pass any number of named arguments.

        Return:
            Result object with build information.
        """
        synth_only = synth_only or self.is_netlist_build

        if output_path is None and not synth_only:
            raise ValueError("Must specify output_path when doing an implementation run")

        if synth_only:
            print(f"Synthesizing Vivado project in {project_path}")
        else:
            print(f"Building Vivado project in {project_path}, placing artifacts in {output_path}")

        # Combine to all available generics. Prefer run-time values over static.
        all_generics = copy_and_combine_dicts(self.static_generics, generics)

        # Run index is optional to specify at build-time
        run_index = self.default_run_index if run_index is None else run_index

        # Send all available information to pre- and post build functions. Prefer build-time values
        # over the static arguments.
        all_parameters = copy_and_combine_dicts(self.other_arguments, pre_and_post_build_parameters)
        all_parameters.update(
            project_path=project_path,
            output_path=output_path,
            run_index=run_index,
            generics=all_generics,
            synth_only=synth_only,
            from_impl=from_impl,
            num_threads=num_threads,
        )

        # The pre-build hooks (either project pre-build hook or any of the module's pre-build hooks)
        # might have side effects. E.g. change some register constants. So we make a deep copy of
        # the module list before any of these hooks are called. Note that the modules are copied
        # before the pre-create hook as well, since we do not know if we might be performing a
        # create-only or build-only operation. The copy does not take any significant time, so this
        # is not an issue.
        self.modules = deepcopy(self.modules)

        result = BuildResult(self.name)

        for module in self.modules:
            if not module.pre_build(project=self, **all_parameters):
                print(
                    f"ERROR: Module {module.name} pre-build hook returned False. Failing the build."
                )
                result.success = False
                return result

            # Make sure register packages are up to date
            module.create_register_synthesis_files()

        if not self.pre_build(**all_parameters):
            print("ERROR: Project pre-build hook returned False. Failing the build.")
            result.success = False
            return result

        # We ignore the type of 'output_path' going from 'Path | None' to 'Path'.
        # It is only used if 'synth_only' is False, and we have an assertion that 'output_path' is
        # not None in that case above.

        build_vivado_project_tcl = self._build_tcl(
            project_path=project_path,
            output_path=output_path,  # type: ignore[arg-type]
            num_threads=num_threads,
            run_index=run_index,
            all_generics=all_generics,
            synth_only=synth_only,
            from_impl=from_impl,
            impl_explore=self.impl_explore,
        )

        if not run_vivado_tcl(self._vivado_path, build_vivado_project_tcl):
            result.success = False
            return result

        result.synthesis_size = self._get_size(project_path, f"synth_{run_index}")
        if self.report_logic_level_distribution:
            result.logic_level_distribution = self._get_logic_level_distribution(
                project_path, f"synth_{run_index}"
            )

        if not synth_only:
            if self.impl_explore:
                runs_path = project_path / f"{self.name}.runs"
                for run in runs_path.iterdir():
                    if "impl_explore_" in run.resolve().name:
                        # Check files for existence, since not all runs may have completed
                        bit_file = run / f"{self.top}.bit"
                        bin_file = run / f"{self.top}.bin"
                        if bit_file.exists() or bin_file.exists():
                            impl_folder = run
                            run_name = run.resolve().name
                            break
            else:
                run_name = f"impl_{run_index}"
                impl_folder = project_path / f"{self.name}.runs" / run_name
                bit_file = impl_folder / f"{self.top}.bit"
                bin_file = impl_folder / f"{self.top}.bin"

            shutil.copy2(bit_file, output_path / f"{self.name}.bit")  # type: ignore[operator]
            shutil.copy2(bin_file, output_path / f"{self.name}.bin")  # type: ignore[operator]
            result.implementation_size = self._get_size(project_path, run_name)

        # Send the result object, along with everything else, to the post-build function
        all_parameters.update(build_result=result)

        if not self.post_build(**all_parameters):
            print("ERROR: Project post-build hook returned False. Failing the build.")
            result.success = False

        return result

    def open(self, project_path: Path) -> bool:
        """
        Open the project in Vivado GUI.

        Arguments:
            project_path: A path containing a Vivado project.

        Return:
            True if everything went well.
        """
        return run_vivado_gui(self._vivado_path, self.project_file(project_path))

    def _get_size(self, project_path: Path, run: str) -> dict[str, int]:
        """
        Reads the hierarchical utilization report and returns the top level size
        for the specified run.
        """
        report_as_string = read_file(
            project_path / f"{self.name}.runs" / run / "hierarchical_utilization.rpt"
        )
        return HierarchicalUtilizationParser.get_size(report_as_string)

    def _get_logic_level_distribution(self, project_path: Path, run: str) -> str:
        """
        Reads the hierarchical utilization report and returns the top level size
        for the specified run.
        """
        report_as_string = read_file(
            project_path / f"{self.name}.runs" / run / "logical_level_distribution.rpt"
        )
        return LogicLevelDistributionParser.get_table(report_as_string)

    def __str__(self) -> str:
        result = f"{self.name}\n"

        if self.defined_at is not None:
            result += f"Defined at: {self.defined_at.resolve()}\n"

        result += f"Type:       {self.__class__.__name__}\n"
        result += f"Top level:  {self.top}\n"

        if self.static_generics:
            generics = self._dict_to_string(self.static_generics)
        else:
            generics = "-"
        result += f"Generics:   {generics}\n"

        if self.other_arguments:
            result += f"Arguments:  {self._dict_to_string(self.other_arguments)}\n"

        return result

    @staticmethod
    def _dict_to_string(data: dict[str, Any]) -> str:
        return ", ".join([f"{name}={value}" for name, value in data.items()])


class VivadoNetlistProject(VivadoProject):
    """
    Used for handling Vivado build of a module without top level pinning.
    """

    def __init__(
        self,
        analyze_synthesis_timing: bool = False,
        build_result_checkers: Optional[list[Union["SizeChecker", "MaximumLogicLevel"]]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Arguments:
            analyze_synthesis_timing: Enable analysis of the synthesized design's timing.
                This will make the build flow open the design, and check for unhandled clock
                crossings and pulse width violations.
                Enabling it will add significant build time (can be as much as +40%).
                Also, in order for clock crossing check to work, the clocks have to be created
                using a constraint file.
            build_result_checkers:
                Checkers that will be executed after a successful build. Is used to automatically
                check that e.g. resource utilization is not greater than expected.
            kwargs: Further arguments accepted by :meth:`.VivadoProject.__init__`.
        """
        super().__init__(**kwargs)

        self.is_netlist_build = True
        self.analyze_synthesis_timing = analyze_synthesis_timing
        self.report_logic_level_distribution = True
        self.build_result_checkers = [] if build_result_checkers is None else build_result_checkers

    def build(  # type: ignore  # pylint: disable=arguments-differ
        self, **kwargs: Any
    ) -> BuildResult:
        """
        Build the project.

        Arguments:
            kwargs: All arguments as accepted by :meth:`.VivadoProject.build`.
        """
        result = super().build(**kwargs)
        result.success = result.success and self._check_size(result)

        return result

    def _check_size(self, build_result: BuildResult) -> bool:
        if not build_result.success:
            print(f"Can not do post_build check for {self.name} since it did not succeed.")
            return False

        success = True
        for build_result_checker in self.build_result_checkers:
            checker_result = build_result_checker.check(build_result)
            success = success and checker_result

        return success


class VivadoIpCoreProject(VivadoProject):
    """
    A Vivado project that is only used to generate simulation models of IP cores.
    """

    ip_cores_only = True

    def __init__(self, **kwargs: Any) -> None:
        """
        Arguments:
            kwargs: Arguments as accepted by :meth:`.VivadoProject.__init__`.
        """
        super().__init__(**kwargs)

    def build(self, **kwargs: Any):  # type: ignore  # pylint: disable=arguments-differ
        """
        Not implemented.
        """
        raise NotImplementedError("IP core project can not be built")


def copy_and_combine_dicts(
    dict_first: Optional[dict[str, Any]], dict_second: Optional[dict[str, Any]]
) -> dict[str, Any]:
    """
    Will prefer values in the second dict, in case the same key occurs in both.
    Will return an empty dictionary if both are ``None``.
    """
    if dict_first is None:
        if dict_second is None:
            return dict()

        return dict_second.copy()

    if dict_second is None:
        return dict_first.copy()

    result = dict_first.copy()
    result.update(dict_second)

    return result



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

"""
Test a subset of what is tested for commercial, since most of the code is inherited
from the common class.
"""

# Standard libraries
from pathlib import Path
from unittest.mock import MagicMock, call, patch

# Third party libraries
import pytest

# First party libraries
from tsfpga.vivado.simlib import VivadoSimlib


# pylint: disable=redefined-outer-name
@pytest.fixture
def simlib_test(tmp_path):
    class SimlibGhdlTestFixture:
        def __init__(self):
            self.output_path = tmp_path / "simlib"

            self.vivado_simlib = self.get_vivado_simlib()

        def get_vivado_simlib(self, ghdl_version_string="GHDL 0.36 ..."):
            with patch("tsfpga.vivado.simlib_ghdl.run_command", autospec=True) as run_command:
                run_command.return_value.stdout = ghdl_version_string

                simulator_class = MagicMock()
                simulator_class.name = "ghdl"
                simulator_class.find_prefix.return_value = "/usr/bin"

                vunit_proj = MagicMock()
                vunit_proj._simulator_class = simulator_class  # pylint: disable=protected-access

                vivado_simlib = VivadoSimlib.init(
                    self.output_path, vunit_proj, Path("/tools/xilinx/Vivado/2019.2/bin/vivado")
                )

                return vivado_simlib

        def assert_should_compile(self):
            assert self.vivado_simlib.compile_is_needed
            with patch(
                "tsfpga.vivado.simlib_ghdl.VivadoSimlibGhdl._compile", autospec=True
            ) as mock:
                self.vivado_simlib.compile_if_needed()
                mock.assert_called_once()

        def assert_should_not_compile(self):
            assert not self.vivado_simlib.compile_is_needed
            with patch(
                "tsfpga.vivado.simlib_ghdl.VivadoSimlibGhdl._compile", autospec=True
            ) as mock:
                self.vivado_simlib.compile_if_needed()
                mock.assert_not_called()

    return SimlibGhdlTestFixture()


def test_should_not_recompile(simlib_test):
    simlib_test.assert_should_compile()
    simlib_test.assert_should_not_compile()


def test_ghdl_version_string(simlib_test):
    assert (
        ".ghdl_0_36_v0_36."
        in simlib_test.get_vivado_simlib(
            ghdl_version_string="GHDL 0.36 (v0.36) [Dunoon edition]"
        ).artifact_name
    )
    assert (
        ".ghdl_0_36."
        in simlib_test.get_vivado_simlib(
            ghdl_version_string="GHDL 0.36 [Dunoon edition]"
        ).artifact_name
    )
    assert (
        ".ghdl_0_36_v0_36."
        in simlib_test.get_vivado_simlib(ghdl_version_string="GHDL 0.36 (v0.36)").artifact_name
    )
    assert (
        ".ghdl_0_36" in simlib_test.get_vivado_simlib(ghdl_version_string="GHDL 0.36").artifact_name
    )

    assert (
        ".ghdl_0_37_dev_v0_36_1605_ge4aa89cd"
        in simlib_test.get_vivado_simlib(
            ghdl_version_string="GHDL 0.37-dev (v0.36-1605-ge4aa89cd) [Dunoon edition]"
        ).artifact_name
    )
    assert (
        ".ghdl_0_37_dev_v0_36_1605_ge4aa89cd."
        in simlib_test.get_vivado_simlib(
            ghdl_version_string="GHDL 0.37-dev (v0.36-1605-ge4aa89cd)"
        ).artifact_name
    )


def test_should_compile_file_by_file_on_windows_but_not_on_linux(simlib_test):
    library_name = "unisim"

    # pylint: disable=protected-access
    unisim_path = simlib_test.vivado_simlib._libraries_path / library_name
    vhd_files = [unisim_path / "a.vhd", unisim_path / "b.vhd"]

    def run_test(is_windows, expected_calls):
        with patch.object(
            simlib_test.vivado_simlib, "_execute_ghdl", autospec=True
        ) as execute_ghdl, patch(
            "tsfpga.vivado.simlib_ghdl.system_is_windows", autospec=True
        ) as system_is_windows:
            system_is_windows.return_value = is_windows

            simlib_test.vivado_simlib._compile_ghdl(vhd_files=vhd_files, library_name=library_name)

            assert execute_ghdl.call_args_list == expected_calls

    def get_expected_call(files):
        return call(
            workdir=simlib_test.vivado_simlib.output_path / library_name,
            library_name=library_name,
            files=files,
        )

    # One call with many files on e.g. Linux.
    expected_calls = [
        get_expected_call(
            files=[
                str(vhd_files[0]),
                str(vhd_files[1]),
            ]
        )
    ]
    run_test(is_windows=False, expected_calls=expected_calls)

    # Many calls with individual file on Windows.
    expected_calls = [
        get_expected_call(files=[str(vhd_files[0])]),
        get_expected_call(files=[str(vhd_files[1])]),
    ]
    run_test(is_windows=True, expected_calls=expected_calls)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Third party libraries
import pytest

# First party libraries
from tsfpga.math_utils import (
    to_binary_nibble_string,
    to_binary_string,
    to_hex_byte_string,
    to_hex_string,
)


def test_to_binary_string():
    assert to_binary_string(value=0, result_width=1) == "0"
    assert to_binary_string(value=1, result_width=1) == "1"
    assert to_binary_string(value=2, result_width=3) == "010"
    assert to_binary_string(value=15, result_width=4) == "1111"
    assert to_binary_string(value=0, result_width=5) == "00000"


def test_to_binary_string_width_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_binary_string(value=15, result_width=0)
    assert str(exception_info.value) == 'Invalid result width "0".'

    with pytest.raises(ValueError) as exception_info:
        to_binary_string(value=-7, result_width=-3)
    assert str(exception_info.value) == 'Invalid result width "-3".'


def test_to_binary_string_value_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_binary_string(value=15, result_width=3)
    assert str(exception_info.value) == 'Value "15" out of 3-bit range.'

    with pytest.raises(ValueError) as exception_info:
        to_binary_string(value=-7, result_width=10)
    assert str(exception_info.value) == 'Value "-7" out of 10-bit range.'


def test_to_binary_nibble_string():
    assert to_binary_nibble_string(value=0, result_width_bits=1) == "0"
    assert to_binary_nibble_string(value=1, result_width_bits=1) == "1"

    assert to_binary_nibble_string(value=2, result_width_bits=3) == "010"
    assert to_binary_nibble_string(value=7, result_width_bits=3) == "111"

    assert to_binary_nibble_string(value=15, result_width_bits=4) == "1111"

    assert to_binary_nibble_string(value=0, result_width_bits=5) == "0_0000"
    assert to_binary_nibble_string(value=16, result_width_bits=5) == "1_0000"

    assert to_binary_nibble_string(value=265, result_width_bits=9) == "1_0000_1001"
    assert to_binary_nibble_string(value=565, result_width_bits=10) == "10_0011_0101"
    assert to_binary_nibble_string(value=565, result_width_bits=12) == "0010_0011_0101"
    assert to_binary_nibble_string(value=2345, result_width_bits=12) == "1001_0010_1001"

    assert to_binary_nibble_string(value=5345, result_width_bits=13) == "1_0100_1110_0001"


def test_to_binary_nibble_string_width_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_binary_nibble_string(value=15, result_width_bits=0)
    assert str(exception_info.value) == 'Invalid result width "0".'

    with pytest.raises(ValueError) as exception_info:
        to_binary_nibble_string(value=-7, result_width_bits=-3)
    assert str(exception_info.value) == 'Invalid result width "-3".'


def test_to_binary_nibble_string_value_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_binary_nibble_string(value=15, result_width_bits=3)
    assert str(exception_info.value) == 'Value "15" out of 3-bit range.'

    with pytest.raises(ValueError) as exception_info:
        to_binary_nibble_string(value=-7, result_width_bits=10)
    assert str(exception_info.value) == 'Value "-7" out of 10-bit range.'


def test_to_hex_string():
    assert to_hex_string(value=0, result_width_bits=1) == "0"
    assert to_hex_string(value=1, result_width_bits=1) == "1"
    assert to_hex_string(value=2, result_width_bits=3) == "2"
    assert to_hex_string(value=2, result_width_bits=4) == "2"
    assert to_hex_string(value=2, result_width_bits=5) == "02"
    assert to_hex_string(value=15, result_width_bits=4) == "F"
    assert to_hex_string(value=0, result_width_bits=5) == "00"


def test_to_hex_string_width_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_hex_string(value=15, result_width_bits=0)
    assert str(exception_info.value) == 'Invalid result width "0".'

    with pytest.raises(ValueError) as exception_info:
        to_hex_string(value=-7, result_width_bits=-3)
    assert str(exception_info.value) == 'Invalid result width "-3".'


def test_to_hex_string_value_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_hex_string(value=15, result_width_bits=3)
    assert str(exception_info.value) == 'Value "15" out of 3-bit range.'

    with pytest.raises(ValueError) as exception_info:
        to_hex_string(value=-7, result_width_bits=10)
    assert str(exception_info.value) == 'Value "-7" out of 10-bit range.'


def test_to_hex_byte_string():
    assert to_hex_byte_string(value=0, result_width_bits=1) == "0"
    assert to_hex_byte_string(value=1, result_width_bits=1) == "1"

    assert to_hex_byte_string(value=2, result_width_bits=3) == "2"
    assert to_hex_byte_string(value=7, result_width_bits=3) == "7"

    assert to_hex_byte_string(value=15, result_width_bits=4) == "F"

    assert to_hex_byte_string(value=0, result_width_bits=5) == "00"
    assert to_hex_byte_string(value=15, result_width_bits=5) == "0F"
    assert to_hex_byte_string(value=16, result_width_bits=5) == "10"

    assert to_hex_byte_string(value=265, result_width_bits=9) == "1_09"
    assert to_hex_byte_string(value=565, result_width_bits=10) == "2_35"
    assert to_hex_byte_string(value=565, result_width_bits=12) == "2_35"
    assert to_hex_byte_string(value=2345, result_width_bits=12) == "9_29"

    assert to_hex_byte_string(value=5345, result_width_bits=13) == "14_E1"


def test_to_hex_byte_string_width_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_hex_byte_string(value=15, result_width_bits=0)
    assert str(exception_info.value) == 'Invalid result width "0".'

    with pytest.raises(ValueError) as exception_info:
        to_hex_byte_string(value=-7, result_width_bits=-3)
    assert str(exception_info.value) == 'Invalid result width "-3".'


def test_to_hex_byte_string_value_out_of_range_should_raise_exception():
    with pytest.raises(ValueError) as exception_info:
        to_hex_byte_string(value=15, result_width_bits=3)
    assert str(exception_info.value) == 'Value "15" out of 3-bit range.'

    with pytest.raises(ValueError) as exception_info:
        to_hex_byte_string(value=-7, result_width_bits=10)
    assert str(exception_info.value) == 'Value "-7" out of 10-bit range.'



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# First party libraries
from tsfpga.vivado.hierarchical_utilization_parser import HierarchicalUtilizationParser


def test_utilization_parser():
    report = """
Writing top or top) or (top here should not cause issues
Also writing pipes | | | | should be fine
Random sections of dashes should be fine:
---------------------------------------------------------------
-----------------------+----------------------------+----------
But this should hit:
| x | x | a | b | c | d |
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
| x | (top) | 1 | 2 | 3 | 4 |
"""

    result = HierarchicalUtilizationParser.get_size(report)
    assert result["a"] == 1
    assert result["b"] == 2
    assert result["c"] == 3
    assert result["d"] == 4
    assert len(result) == 4



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import re


class LogicLevelDistributionParser:
    """
    Used for parsing the ``report_design_analysis -logic_level_distribution`` report generated
    by Vivado.

    This code is very hard coded for how the report and table is formatted. See the unit tests
    for examples of this formatting.
    """

    @staticmethod
    def get_table(report: str) -> str:
        """
        Takes a report as a string and returns the table.

        Arguments:
            report: A string containing the entire Vivado
                ``report_design_analysis -logic_level_distribution`` report.
        """
        table_regexp = re.compile(r"\n(\+--.*-\+)\n", re.DOTALL)
        match = table_regexp.search(report)
        if match:
            return match.group(1)

        raise ValueError(f"Could not find table in report: {report}")

    @staticmethod
    def get_maximum_logic_level(table: str) -> int:
        """
        Returns the maximum logic level in the table.

        Arguments:
            table: The table as returned by :meth:`.get_table`.

        Return:
            The maximum logic level.
        """
        header_line = table.split("\n")[1]
        # First and last items are empty, due to leading and trailing "|" in the table
        headings = header_line.split("|")

        if len(headings) <= 4:
            # Nothing in table. Happens if there are no paths in the design.
            # 4 elements in list means two headings, i.e. no integer levels.
            # See unit test for examples of how this looks.
            return 0

        right_most_heading = headings[-2]
        return int(right_most_heading.strip())



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# First party libraries
from tsfpga.system_utils import read_file


def file_contains_string(file, string):
    return string in read_file(file)


def file_equals(file, string):
    return string == read_file(file)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import re
import subprocess
from pathlib import Path
from typing import Iterable, Optional, Union

# Local folder libraries
from .system_utils import file_is_in_directory, run_command


def get_svn_revision_information(cwd: Optional[Path] = None) -> str:
    """
    Get a string describing the current SVN commit.
    E.g. ``"r1234"`` or ``"r1234 (local changes present)"``.

    Arguments:
        cwd: The directory where SVN commands will be run.
    """
    check_that_svn_commands_are_available(cwd=cwd)

    result = f"r{get_svn_revision(cwd=cwd)}"

    if svn_local_changes_are_present(cwd=cwd):
        result += " (local changes present)"

    return result


def svn_commands_are_available(cwd: Optional[Path] = None) -> bool:
    """
    True if "svn" command executable is available and ``cwd`` is in a valid SVN repo.

    Arguments:
        cwd: The directory where SVN commands will be run.
    """
    try:
        get_svn_revision(cwd=cwd)
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False
    return True


def check_that_svn_commands_are_available(cwd: Optional[Path] = None) -> None:
    """
    Raise an exception unless "svn" command executable is available and ``cwd`` is in a valid
    SVN repo.

    Arguments:
        cwd: The directory where SVN commands will be run.
    """
    if not svn_commands_are_available(cwd):
        message = (
            "Could not run SVN. Is the command available on PATH? Is the script called from a repo?"
        )
        raise RuntimeError(message)


def get_svn_revision(cwd: Optional[Path] = None) -> int:
    """
    Get the current SVN revision number.

    Arguments:
        cwd: The directory where SVN commands will be run.
    """
    command = ["svn", "info", "--show-item", "revision"]
    stdout: str = run_command(cmd=command, cwd=cwd, capture_output=True).stdout

    # Remove trailing newline
    return int(stdout.strip())


def svn_local_changes_are_present(cwd: Optional[Path] = None) -> bool:
    """
    Return true if the repo contains changes that have been made after the last commit.
    Info from here: https://rubyinrails.com/2014/01/11/svn-command-to-check-modified-files/

    Arguments:
        cwd: The directory where SVN commands will be run.
    """
    command = ["svn", "status"]
    stdout: str = run_command(cmd=command, cwd=cwd, capture_output=True).stdout

    # Status code for file Added, Deleted, Modified, in Conflict or missing
    regexp = re.compile(r"\n[ADMC!] ")
    return regexp.search(stdout) is not None


RE_SVN_STATUS_LINE = re.compile(r"^.+\d+\s+\d+\s+\S+\s+(\S+)$")


def find_svn_files(
    directory: Path,
    excludes: Optional[list[Path]] = None,
    file_endings_include: Optional[Union[str, tuple[str, ...]]] = None,
    file_endings_avoid: Optional[Union[str, tuple[str, ...]]] = None,
) -> Iterable[Path]:
    """
    Find files that are checked in to SVN. It runs "svn status" rather than "svn ls". This means
    that it is a local operation, that does not require credentials or any connection with
    an SVN server.

    Arguments:
        directory: Search in this directory.
        excludes: These files and folders will not be included.
        file_endings_include: Only files with these endings will be included.
        file_endings_avoid: Files with these endings will not be included.

    Return:
        The files that are available in SVN.
    """
    excludes = [] if excludes is None else [exclude.resolve() for exclude in excludes]

    command = ["svn", "status", "-v"]
    stdout = run_command(cmd=command, cwd=directory, capture_output=True).stdout

    for line in stdout.split("\n"):
        match = RE_SVN_STATUS_LINE.match(line)
        if not match:
            continue

        svn_file = match.group(1)
        file_path = directory / svn_file

        # Make sure concatenation of relative paths worked
        assert file_path.exists(), file_path

        if file_path.is_dir():
            continue

        if file_endings_include is not None and not file_path.name.endswith(file_endings_include):
            continue

        if file_endings_avoid is not None and file_path.name.endswith(file_endings_avoid):
            continue

        if file_is_in_directory(file_path, excludes):
            continue

        yield file_path



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

"""
Import this file in order to have the default locations of the hdl-registers/hdl-registers
and vunit/vunit repos added to PYTHONPATH.
"""

# Standard libraries
import sys

# First party libraries
import tsfpga

# Do PYTHONPATH insert() instead of append() to prefer any local repo checkout over any pip install

# Paths e.g.
# repo/tsfpga/tsfpga
# repo/hdl-registers/hdl-registers
PATH_TO_HDL_REGISTERS = tsfpga.REPO_ROOT.parent.parent.resolve() / "hdl-registers" / "hdl-registers"
sys.path.insert(0, str(PATH_TO_HDL_REGISTERS))

# Paths e.g.
# repo/tsfpga/tsfpga
# repo/vunit/vunit
PATH_TO_VUNIT = tsfpga.REPO_ROOT.parent.parent.resolve() / "vunit" / "vunit"
sys.path.insert(0, str(PATH_TO_VUNIT))



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import platform
import zipfile
from abc import ABC, abstractmethod
from pathlib import Path
from shutil import make_archive
from typing import Optional

# First party libraries
from tsfpga.system_utils import create_file, delete
from tsfpga.vivado.common import get_vivado_version

# Local folder libraries
from .common import get_vivado_path


class VivadoSimlibCommon(ABC):
    """
    Class for handling Vivado simlib used for simulation. Keeps track of when a
    (re)compile is needed.

    This is a base class that defines an interface and some common methods.
    See subclasses for details: :class:`.VivadoSimlibGhdl`, :class:`.VivadoSimlibCommercial`.
    """

    # The version of this class. Can be bumped to force a re-compile if e.g. the TCL script changes
    # or the output folder structure is updated.
    _format_version_id = 4

    # Set in subclass to a list of strings. The libraries that shall be compiled and added to
    # VUnit project.
    library_names: list[str]

    def __init__(self, vivado_path: Optional[Path], output_path: Path) -> None:
        """
        Call from subclass. Please do not instantiate this class directly.
        """
        self._vivado_path = get_vivado_path(vivado_path)
        self._libraries_path = (self._vivado_path.parent.parent / "data" / "vhdl" / "src").resolve()

        self.output_path = output_path.resolve() / self._get_version_tag()

    def compile_if_needed(self) -> bool:
        """
        Compile if needed (if :meth:`compile_is_needed <.compile_is_needed>` condition is not
        fulfilled).
        """
        if self.compile_is_needed:
            self.compile()
            return True

        return False

    @property
    def compile_is_needed(self) -> bool:
        """
        If there is compiled simlib available that matches

         * Operating system
         * Vivado version
         * Simulator version

        then there should not be a recompile.

        .. note::
            Subclass implementations might add further conditions.

        Return:
            True if compiled simlib is not available. False otherwise.
        """
        if self._done_token.exists():
            return False

        return True

    def compile(self) -> None:
        """
        Compile simlib.
        """
        # Probably does not exists, but try to delete just in case.
        delete(self._done_token)

        print(f"Compiling Vivado simlib from {self._libraries_path} into {self.output_path}...")
        self._compile()

        create_file(self._done_token, "Done!")

    @abstractmethod
    def _compile(self) -> None:
        """
        Compile simlib.
        Overload in a subclass.
        """

    def add_to_vunit_project(self) -> None:
        """
        Add the compiled simlib to your VUnit project.
        """
        self._add_to_vunit_project()

    @abstractmethod
    def _add_to_vunit_project(self) -> None:
        """
        Add the compiled simlib to your VUnit project.
        Overload in a subclass.
        """

    @property
    def artifact_name(self) -> str:
        """
        The name of the folder where simlib is or will be compiled.
        Follows a format ``vivado-simlib-WW.XX.YY.ZZ`` suitable for storage and versioning
        in Artifactory.
        """
        return self.output_path.name

    def to_archive(self) -> Path:
        """
        Compress compiled simlib to an archive.

        Return:
            Path to the archive.
        """
        make_archive(str(self.output_path), "zip", self.output_path)
        archive = self.output_path.parent / (self.output_path.name + ".zip")

        return archive

    def from_archive(self, archive: Path) -> None:
        """
        Unpack compiled simlib from an existing archive.

        Arguments:
            archive (pathlib.Path): Path to a zip archive with previously compiled simlib.
        """
        with zipfile.ZipFile(archive, "r") as zip_handle:
            zip_handle.extractall(self.output_path)

    def _get_version_tag(self) -> str:
        tag = "vivado-simlib-"
        tag += self._get_operating_system_tag()
        tag += f".{self._get_vivado_version_tag()}"
        tag += f".{self._get_simulator_tag()}"
        tag += f".format_{self._format_version_id}"

        return tag

    def _get_operating_system_tag(self) -> str:
        """
        Return e.g. "linux".
        """
        return self._format_version(platform.system())

    def _get_vivado_version_tag(self) -> str:
        """
        Return e.g. "vivado_2021_2".
        """
        vivado_version = get_vivado_version(self._vivado_path)

        return self._format_version(f"vivado_{vivado_version}")

    @abstractmethod
    def _get_simulator_tag(self) -> str:
        """
        Return a simulator tag as a string, e.g. "ghdl_1.2.3".
        Overload in a subclass.
        """

    @staticmethod
    def _format_version(version: str) -> str:
        """
        Format version string to something suitable for artifactory versioning.
        """
        return version.replace(".", "_").replace("-", "_").lower()

    @property
    def _done_token(self) -> Path:
        """
        Path to "done" token file.
        """
        return self.output_path / "done.txt"



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# First party libraries
from tsfpga.module import BaseModule


class Module(BaseModule):
    def get_simulation_files(  # type: ignore  # pylint: disable=arguments-differ
        self, include_ip_cores: bool, **kwargs
    ):
        """
        Exclude files that depend on IP cores, if instructed to by the simulation script.
        """
        files_that_depend_on_ip_cores = {
            self.path / "src" / "multiplication.vhd",
            self.path / "test" / "tb_multiplication.vhd",
        }
        files_avoid = None if include_ip_cores else files_that_depend_on_ip_cores

        return super().get_simulation_files(files_avoid=files_avoid, **kwargs)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------
# A set of tools for versions and releases. Should be reusable between projects.
# --------------------------------------------------------------------------------------------------

# Standard libraries
import json
import re
import sys
from pathlib import Path
from urllib.request import urlopen

# Third party libraries
from git.repo import Repo
from packaging.version import Version, parse

# First party libraries
from tsfpga.system_utils import create_file, read_file

# Contents of unreleased.rst when it is empty
UNRELEASED_EMPTY = "Nothing here yet.\n"


class VersionNumberHandler:
    """
    Class for handling the version number in __init__.py.
    """

    _version_regexp = re.compile(r"\n__version__ = \"(\S+?)\"\n")

    def __init__(self, repo: Repo, version_file_path: Path) -> None:
        """
        Arguments:
            repo: The git repository to work with.
            version_file_path: The ``__init__.py`` file that shall be modified.
        """
        self._repo = repo
        self._file_path = version_file_path

    def update(self, new_version: str) -> None:
        """
        Set a new version number.

        Arguments:
            new_version: New version number as a string, e.g. "2.3.1".
        """
        old_version = self._get_current_version()
        self._verify_that_newer_version_number_is_greater_than_older(
            older_version=old_version, newer_version=new_version
        )

        self._set_new_version(new_version)

    def bump_to_prelease(self) -> None:
        """
        Bump to next version number. E.g. go from 8.0.2 to 8.0.3-dev.
        """
        current_version = self._get_current_version()
        (current_major, current_minor, current_patch) = current_version.release

        new_version = f"{current_major}.{current_minor}.{current_patch + 1}-dev"
        self._verify_that_newer_version_number_is_greater_than_older(
            older_version=current_version, newer_version=new_version
        )

        self._set_new_version(new_version)

    @staticmethod
    def _verify_that_newer_version_number_is_greater_than_older(
        older_version: Version, newer_version: str
    ) -> None:
        if older_version >= parse(newer_version):
            sys.exit(f"New version {newer_version} is not greater than old version {older_version}")

    def _get_current_version(self) -> Version:
        data = read_file(self._file_path)

        match = self._version_regexp.search(data)
        if match is None:
            raise RuntimeError(f"Could not find version value in {self._file_path}")

        version = match.group(1)
        return parse(version)

    def _set_new_version(self, new_version: str) -> None:
        data = read_file(self._file_path)

        updated_data = self._version_regexp.sub(f'\n__version__ = "{new_version}"\n', data)
        create_file(self._file_path, updated_data)

        # Add file so that it gets included in upcoming commit
        self._repo.index.add(str(self._file_path.resolve()))


def verify_new_version_number(
    repo: Repo, pypi_project_name: str, new_version: str, unreleased_notes_file: Path
) -> str:
    """
    Verify that the new version number is sane for this project. Will check git log and PyPI
    release history.

    Arguments:
        repo: The git repository to work with.
        pypi_project_name: The name of this project on PyPI.
        version: New version.
        unreleased_notes_file: Path to the "unreleased.rst" release notes file.
            Must not be empty.

    Return:
        The name of the git tag that corresponds to the new version number.
    """
    if repo.is_dirty():
        sys.exit("Must make release from clean repo")

    if read_file(unreleased_notes_file) in ["", UNRELEASED_EMPTY]:
        sys.exit(f"The unreleased notes file {unreleased_notes_file} should not be empty")

    with urlopen(f"https://pypi.python.org/pypi/{pypi_project_name}/json") as file_handle:
        json_data = json.load(file_handle)
        if new_version in json_data["releases"]:
            sys.exit(f"Release {new_version} already exists in PyPI")

    git_tag = f"v{new_version}"
    if git_tag in repo.tags:
        sys.exit(f"Git release tag already exists: {git_tag}")

    return git_tag


def commit_and_tag_release(repo: Repo, version: str, git_tag: str) -> None:
    """
    Make a git commit with a "release" message, and tag it.

    Arguments:
        repo: The git repository to work with.
        version: New version.
        git_tag: New git tag.
    """
    make_commit(repo=repo, commit_message=f"Release version {version}")

    repo.create_tag(git_tag)
    if git_tag not in repo.tags:
        sys.exit("Git tag failed")


def make_commit(repo: Repo, commit_message: str) -> None:
    """
    Make a git commit, and check that it worked.

    Arguments:
        repo (`git.Repo`): The git repository to work with.
        commit_message (str): Commit message to use.
    """
    repo.index.commit(commit_message)
    if repo.is_dirty():
        sys.exit("Git commit failed")



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import argparse
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    # First party libraries
    from tsfpga.module_list import ModuleList

# Do PYTHONPATH insert() instead of append() to prefer any local repo checkout over any pip install
REPO_ROOT = Path(__file__).parent.parent.parent.resolve()
sys.path.insert(0, str(REPO_ROOT))

# Import before others since it modifies PYTHONPATH. pylint: disable=unused-import
import tsfpga.examples.example_pythonpath  # noqa: F401

# First party libraries
import tsfpga
import tsfpga.create_vhdl_ls_config
from tsfpga.create_ghdl_ls_config import create_ghdl_ls_configuration
from tsfpga.examples.example_env import (
    TSFPGA_EXAMPLES_TEMP_DIR,
    get_hdl_modules,
    get_tsfpga_example_modules,
)
from tsfpga.examples.simulation_utils import (
    SimulationProject,
    create_vhdl_ls_configuration,
    get_arguments_cli,
)
from tsfpga.git_simulation_subset import GitSimulationSubset


def main() -> None:
    """
    Main function for the simulation flow. If you are setting up a new simulation environment
    you probably want to copy and modify this function. The other functions and classes
    should be reusable in most cases.
    """
    cli = get_arguments_cli(default_output_path=TSFPGA_EXAMPLES_TEMP_DIR)
    args = cli.parse_args()

    modules = get_tsfpga_example_modules()

    # Avoid the module that depends on Xilinx unisim library
    module_names_avoid = set(["hard_fifo"]) if args.vivado_skip else None
    modules_no_sim = get_hdl_modules(names_avoid=module_names_avoid)

    if args.vcs_minimal:
        if args.test_patterns != "*":
            sys.exit(
                "Can not specify a test pattern when using the --vcs-minimal flag."
                f" Got {args.test_patterns}",
            )

        test_filters = find_git_test_filters(
            args=args,
            repo_root=tsfpga.REPO_ROOT,
            modules=modules,
            modules_no_sim=modules_no_sim,
            reference_branch="origin/main",
        )
        if not test_filters:
            print("Nothing to run. Appears to be no VHDL-related git diff.")
            return

        # Override the test pattern argument to VUnit
        args.test_patterns = test_filters
        print(f"Running VUnit with test pattern {args.test_patterns}")

        # Enable minimal compilation in VUnit
        args.minimal = True

    simulation_project = SimulationProject(args=args)
    simulation_project.add_modules(
        args=args,
        modules=modules,
        modules_no_sim=modules_no_sim,
        include_verilog_files=False,
        include_systemverilog_files=False,
    )
    simlib = simulation_project.add_vivado_simlib()
    ip_core_vivado_project_directory = simulation_project.add_vivado_ip_cores(
        modules=modules + modules_no_sim
    )

    create_vhdl_ls_configuration(
        output_path=tsfpga.REPO_ROOT,
        temp_files_path=TSFPGA_EXAMPLES_TEMP_DIR,
        modules=modules + modules_no_sim,
        ip_core_vivado_project_directory=ip_core_vivado_project_directory,
    )

    create_ghdl_ls_configuration(
        output_path=tsfpga.REPO_ROOT,
        modules=modules + modules_no_sim,
        vunit_proj=simulation_project.vunit_proj,
        simlib=simlib,
    )

    simulation_project.vunit_proj.main()


def find_git_test_filters(
    args: argparse.Namespace,
    repo_root: Path,
    modules: "ModuleList",
    modules_no_sim: Optional["ModuleList"] = None,
    reference_branch: str = "origin/master",
    **setup_vunit_kwargs: Any,
) -> list[str]:
    """
    Construct a VUnit test filter that will run all test cases that are affected by git changes.
    The current git state is compared to a reference branch, and differences are derived.
    See :class:`.GitSimulationSubset` for details.

    Arguments:
        args: Command line argument namespace.
        repo_root: Path to the repository root. Git commands will be run here.
        modules: Will be passed on to :meth:`.SimulationProject.add_modules`.
        modules_no_sim: Will be passed on to :meth:`.SimulationProject.add_modules`.
        reference_branch (str): The name of the reference branch that is used to collect a diff.
        setup_vunit_kwargs : Will be passed on to :meth:`.SimulationProject.add_modules`.

    Return:
        A list of VUnit test case filters.
    """
    # Set up a dummy VUnit project that will be used for dependency scanning.
    # Note that sources are added identical to the "real" project above.
    simulation_project = SimulationProject(args=args)
    simulation_project.add_modules(
        args=args,
        modules=modules,
        modules_no_sim=modules_no_sim,
        include_verilog_files=False,
        include_systemverilog_files=False,
        **setup_vunit_kwargs,
    )

    testbenches_to_run = GitSimulationSubset(
        repo_root=repo_root,
        reference_branch=reference_branch,
        vunit_proj=simulation_project.vunit_proj,
    ).find_subset()

    test_filters = []
    for testbench_file_name, library_name in testbenches_to_run:
        test_filters.append(f"{library_name}.{testbench_file_name}.*")

    return test_filters


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import re
from collections import OrderedDict


class HierarchicalUtilizationParser:
    """
    Used for parsing the ``report_utilization -hierarchical`` report generated by Vivado.
    """

    @staticmethod
    def get_size(report: str) -> dict[str, int]:
        """
        Takes a hierarchical utilization report as a string and returns the top level size
        for the specified run.

        Arguments:
            report (str): A string containing the entire Vivado hierarchical utilization report.
        """
        lines = report.split("\n")
        for idx, line in enumerate(lines):
            # Find the table line that is the top level
            if re.search(r"\(top\)", line):
                # Parse the report, remove uninteresting fields and create dictionary
                # Note that "|" is the column separator. Heading titles for the data is two lines
                # above the row for the top level.
                headers = [column_data.strip() for column_data in lines[idx - 2].split("|")]
                numbers = [column_data.strip() for column_data in line.split("|")]

                # The first columns contain entity name, etc. We only want the numbers
                headers = headers[3:-1]
                numbers = numbers[3:-1]

                # Convert numbers from string to integers
                numbers_int = [int(number) for number in numbers]

                return OrderedDict(zip(headers, numbers_int))

        return {}



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import random
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, Union

# Third party libraries
from hdl_registers.generator.vhdl.axi_lite.wrapper import VhdlAxiLiteWrapperGenerator
from hdl_registers.generator.vhdl.record_package import VhdlRecordPackageGenerator
from hdl_registers.generator.vhdl.register_package import VhdlRegisterPackageGenerator
from hdl_registers.generator.vhdl.simulation.check_package import (
    VhdlSimulationCheckPackageGenerator,
)
from hdl_registers.generator.vhdl.simulation.read_write_package import (
    VhdlSimulationReadWritePackageGenerator,
)
from hdl_registers.generator.vhdl.simulation.wait_until_package import (
    VhdlSimulationWaitUntilPackageGenerator,
)
from hdl_registers.parser.toml import from_toml
from hdl_registers.register import Register
from hdl_registers.register_list import RegisterList

# First party libraries
from tsfpga.constraint import Constraint
from tsfpga.hdl_file import HdlFile
from tsfpga.ip_core_file import IpCoreFile
from tsfpga.module_list import ModuleList
from tsfpga.system_utils import load_python_module

if TYPE_CHECKING:
    # Local folder libraries
    from .vivado.project import VivadoProject


class BaseModule:
    """
    Base class for handling a HDL module with RTL code, constraints, etc.

    Files are gathered from a lot of different sub-folders, to accommodate for projects having
    different catalog structure.
    """

    # Set to False if you do not want to create these register artifacts for this module.
    # Can be done in a child class or on an object instance.
    # Note that artifacts will only be created if the module actually has any registers.
    create_register_package = True
    create_record_package = True
    create_axi_lite_wrapper = True
    create_simulation_read_write_package = True
    create_simulation_check_package = True
    create_simulation_wait_until_package = True

    def __init__(
        self, path: Path, library_name: str, default_registers: Optional[list[Register]] = None
    ):
        """
        Arguments:
            path: Path to the module folder.
            library_name: VHDL library name.
            default_registers: Default registers.
        """
        self.path = path.resolve()
        self.name = path.name
        self.library_name = library_name

        self._default_registers = default_registers
        self._registers: Optional[RegisterList] = None

    @staticmethod
    def _get_file_list(
        folders: list[Path],
        file_endings: Union[str, tuple[str, ...]],
        files_include: Optional[set[Path]] = None,
        files_avoid: Optional[set[Path]] = None,
    ) -> list[Path]:
        """
        Returns a list of files given a list of folders.

        Arguments:
            folders: The folders to search.
            file_endings: File endings to include.
            files_include: Optionally filter to only include these files.
            files_avoid: Optionally filter to discard these files.
        """
        files = []
        for folder in folders:
            for file in folder.glob("*"):
                if not file.is_file():
                    continue

                if not file.name.lower().endswith(file_endings):
                    continue

                if files_include is not None and file not in files_include:
                    continue

                if files_avoid is not None and file in files_avoid:
                    continue

                files.append(file)

        return files

    def _get_hdl_file_list(  # pylint: disable=too-many-arguments
        self,
        folders: list[Path],
        files_include: Optional[set[Path]] = None,
        files_avoid: Optional[set[Path]] = None,
        include_vhdl_files: bool = True,
        include_verilog_files: bool = True,
        include_systemverilog_files: bool = True,
    ) -> list[HdlFile]:
        """
        Return a list of HDL file objects.
        """
        file_endings: tuple[str, ...] = tuple()
        if include_vhdl_files:
            file_endings += HdlFile.file_endings_mapping[HdlFile.Type.VHDL]
        if include_verilog_files:
            file_endings += HdlFile.file_endings_mapping[HdlFile.Type.VERILOG_SOURCE]
            file_endings += HdlFile.file_endings_mapping[HdlFile.Type.VERILOG_HEADER]
        if include_systemverilog_files:
            file_endings += HdlFile.file_endings_mapping[HdlFile.Type.SYSTEMVERILOG_SOURCE]
            file_endings += HdlFile.file_endings_mapping[HdlFile.Type.SYSTEMVERILOG_HEADER]

        return [
            HdlFile(path=file_path)
            for file_path in self._get_file_list(
                folders=folders,
                file_endings=file_endings,
                files_include=files_include,
                files_avoid=files_avoid,
            )
        ]

    @property
    def registers(self) -> Optional[RegisterList]:
        """
        Get the registers for this module.
        Will be ``None`` if the module doesn't have any registers.
        I.e. if no TOML file exists and no hook creates registers.
        """
        if self._registers:
            # Only create object from TOML once.
            return self._registers

        toml_file = self.path / f"regs_{self.name}.toml"
        if toml_file.exists():
            self._registers = from_toml(
                name=self.name, toml_file=toml_file, default_registers=self._default_registers
            )

        self.registers_hook()
        return self._registers

    def registers_hook(self) -> None:
        """
        This function will be called directly after creating this module's registers from
        the TOML definition file.
        If the TOML file does not exist this hook will still be called, but the module's registers
        will be ``None``.

        This is a good place if you want to add or modify some registers from Python.
        Override this method and implement the desired behavior in a subclass.

        .. Note::
            This default method does nothing.
            Shall be overridden by modules that utilize this mechanism.
        """

    def create_register_synthesis_files(self) -> None:
        """
        Create the register artifacts that are needed for synthesis.
        If this module does not have registers, this method does nothing.
        """
        if self.registers is not None:
            # Delete any old file that might exist so we don't have multiple and
            # outdated definitions.
            # This package location was used before the separate register folders were introduced,
            # back when we only created one register artifact.
            old_regs_pkg = self.path / f"{self.name}_regs_pkg.vhd"
            if old_regs_pkg.exists():
                old_regs_pkg.unlink()

            if self.create_register_package:
                VhdlRegisterPackageGenerator(
                    register_list=self.registers, output_folder=self.register_synthesis_folder
                ).create_if_needed()

            if self.create_record_package:
                VhdlRecordPackageGenerator(
                    register_list=self.registers, output_folder=self.register_synthesis_folder
                ).create_if_needed()

            if self.create_axi_lite_wrapper:
                VhdlAxiLiteWrapperGenerator(
                    register_list=self.registers, output_folder=self.register_synthesis_folder
                ).create_if_needed()

    def create_register_simulation_files(self) -> None:
        """
        Create the register artifacts that are needed for simulation.
        Does not create the implementation files, which are also technically needed for simulation.
        So a call to :meth:`.create_register_synthesis_files` must also be done.

        If this module does not have registers, this method does nothing.
        """
        if self.registers is not None:
            if self.create_simulation_read_write_package:
                VhdlSimulationReadWritePackageGenerator(
                    register_list=self.registers, output_folder=self.register_simulation_folder
                ).create_if_needed()

            if self.create_simulation_check_package:
                VhdlSimulationCheckPackageGenerator(
                    register_list=self.registers, output_folder=self.register_simulation_folder
                ).create_if_needed()

            if self.create_simulation_wait_until_package:
                VhdlSimulationWaitUntilPackageGenerator(
                    register_list=self.registers, output_folder=self.register_simulation_folder
                ).create_if_needed()

    @property
    def synthesis_folders(self) -> list[Path]:
        """
        Synthesis/implementation source code files will be gathered from these folders.
        """
        return [
            self.path,
            self.path / "src",
            self.path / "rtl",
            self.path / "hdl" / "rtl",
            self.path / "hdl" / "package",
            self.register_synthesis_folder,
        ]

    @property
    def register_synthesis_folder(self) -> Path:
        """
        Generated register artifacts that are needed for synthesis/implementation will be
        placed in this folder.
        """
        return self.path / "regs_src"

    @property
    def sim_folders(self) -> list[Path]:
        """
        Files with simulation models (the ``sim`` folder) will be gathered from these folders.
        """
        return [
            self.path / "sim",
            self.register_simulation_folder,
        ]

    @property
    def register_simulation_folder(self) -> Path:
        """
        Generated register artifacts that are needed for simulation will be placed in this folder.
        """
        return self.path / "regs_sim"

    @property
    def test_folders(self) -> list[Path]:
        """
        Testbench files will be gathered from these folders.
        """
        return [
            self.path / "test",
            self.path / "rtl" / "tb",
        ]

    def get_synthesis_files(  # pylint: disable=unused-argument
        self,
        files_include: Optional[set[Path]] = None,
        files_avoid: Optional[set[Path]] = None,
        include_vhdl_files: bool = True,
        include_verilog_files: bool = True,
        include_systemverilog_files: bool = True,
        **kwargs: Any,
    ) -> list[HdlFile]:
        """
        Get a list of files that shall be included in a synthesis project.

        The ``files_include`` and ``files_avoid`` arguments can be used to filter what files are
        included.
        This can be useful in many situations, e.g. when encrypted files of files that include an
        IP core shall be avoided.
        It is recommended to overload this function in a subclass in your ``module_*.py``,
        and call this super method with the arguments supplied.

        Arguments:
            files_include: Optionally filter to only include these files.
            files_avoid: Optionally filter to discard these files.
            include_vhdl_files: Optionally disable inclusion of files with VHDL
                file endings.
            include_verilog_files: Optionally disable inclusion of files with Verilog
                file endings.
            include_systemverilog_files: Optionally disable inclusion of files with SystemVerilog
                file endings.
            kwargs: Further parameters that can be sent by build flow to control what
                files are included.

        Return:
            Files that should be included in a synthesis project.
        """
        self.create_register_synthesis_files()

        return self._get_hdl_file_list(
            folders=self.synthesis_folders,
            files_include=files_include,
            files_avoid=files_avoid,
            include_vhdl_files=include_vhdl_files,
            include_verilog_files=include_verilog_files,
            include_systemverilog_files=include_systemverilog_files,
        )

    def get_simulation_files(  # pylint: disable=too-many-arguments
        self,
        include_tests: bool = True,
        files_include: Optional[set[Path]] = None,
        files_avoid: Optional[set[Path]] = None,
        include_vhdl_files: bool = True,
        include_verilog_files: bool = True,
        include_systemverilog_files: bool = True,
        **kwargs: Any,
    ) -> list[HdlFile]:
        """
        See :meth:`.get_synthesis_files` for instructions on how to use ``files_include``
        and ``files_avoid``.

        Arguments:
            include_tests: When ``False``, the ``test`` files are not included
                (the ``sim`` files are always included).
            files_include: Optionally filter to only include these files.
            files_avoid: Optionally filter to discard these files.
            include_vhdl_files: Optionally disable inclusion of files with VHDL
                file endings.
            include_verilog_files: Optionally disable inclusion of files with Verilog
                file endings.
            include_systemverilog_files: Optionally disable inclusion of files with SystemVerilog
                file endings.
            kwargs: Further parameters that can be sent by simulation flow to control what
                files are included.

        Return:
            Files that should be included in a simulation project.
        """
        # Shallow copy the list since we might append to it.
        sim_and_test_folders = self.sim_folders.copy()

        if include_tests:
            sim_and_test_folders += self.test_folders

        self.create_register_simulation_files()

        test_files = self._get_hdl_file_list(
            folders=sim_and_test_folders,
            files_include=files_include,
            files_avoid=files_avoid,
            include_vhdl_files=include_vhdl_files,
            include_verilog_files=include_verilog_files,
            include_systemverilog_files=include_systemverilog_files,
        )

        synthesis_files = self.get_synthesis_files(
            files_include=files_include,
            files_avoid=files_avoid,
            include_vhdl_files=include_vhdl_files,
            include_verilog_files=include_verilog_files,
            include_systemverilog_files=include_systemverilog_files,
            **kwargs,
        )

        return synthesis_files + test_files

    def get_documentation_files(  # pylint: disable=unused-argument
        self,
        files_include: Optional[set[Path]] = None,
        files_avoid: Optional[set[Path]] = None,
        include_vhdl_files: bool = True,
        include_verilog_files: bool = True,
        include_systemverilog_files: bool = True,
        **kwargs: Any,
    ) -> list[HdlFile]:
        """
        Get a list of files that shall be included in a documentation build.

        It will return all files from the module except testbenches and any generated
        register package.
        Overwrite in a subclass if you want to change this behavior.

        Arguments:
            files_include: Optionally filter to only include these files.
            files_avoid: Optionally filter to discard these files.
            include_vhdl_files: Optionally disable inclusion of files with VHDL
                file endings.
            include_verilog_files: Optionally disable inclusion of files with Verilog
                file endings.
            include_systemverilog_files: Optionally disable inclusion of files with SystemVerilog
                file endings.

        Return:
            Files that should be included in documentation.
        """
        # Do not include generated register code in the documentation.
        files_to_avoid = set(
            self._get_file_list(
                folders=[self.register_synthesis_folder, self.register_simulation_folder],
                file_endings="vhd",
            )
        )
        if files_avoid:
            files_to_avoid |= files_avoid

        return self._get_hdl_file_list(
            folders=self.synthesis_folders + self.sim_folders,
            files_include=files_include,
            files_avoid=files_to_avoid,
            include_vhdl_files=include_vhdl_files,
            include_verilog_files=include_verilog_files,
            include_systemverilog_files=include_systemverilog_files,
        )

    # pylint: disable=unused-argument
    def get_ip_core_files(
        self,
        files_include: Optional[set[Path]] = None,
        files_avoid: Optional[set[Path]] = None,
        **kwargs: Any,
    ) -> list[IpCoreFile]:
        """
        Get IP cores for this module.

        Note that the :class:`.ip_core_file.IpCoreFile` class accepts a ``variables`` argument that
        can be used to parameterize IP core creation. By overloading this method in a subclass
        you can pass on ``kwargs`` arguments from the build/simulation flow to
        :class:`.ip_core_file.IpCoreFile` creation to achieve this parameterization.

        Arguments:
            files_include: Optionally filter to only include these files.
            files_avoid: Optionally filter to discard these files.
            kwargs: Further parameters that can be sent by build/simulation flow to control what
                IP cores are included and what their variables are.

        Return:
            The IP cores for this module.
        """
        folders = [
            self.path / "ip_cores",
        ]
        file_endings = "tcl"
        return [
            IpCoreFile(ip_core_file)
            for ip_core_file in self._get_file_list(
                folders=folders,
                file_endings=file_endings,
                files_include=files_include,
                files_avoid=files_avoid,
            )
        ]

    # pylint: disable=unused-argument
    def get_scoped_constraints(
        self,
        files_include: Optional[set[Path]] = None,
        files_avoid: Optional[set[Path]] = None,
        **kwargs: Any,
    ) -> list[Constraint]:
        """
        Constraints that shall be applied to a certain entity within this module.

        Arguments:
            files_include: Optionally filter to only include these files.
            files_avoid: Optionally filter to discard these files.
            kwargs: Further parameters that can be sent by build/simulation flow to control what
                constraints are included.

        Return:
            The constraints.
        """
        folders = [
            self.path / "scoped_constraints",
            self.path / "entity_constraints",
            self.path / "hdl" / "constraints",
        ]
        file_endings = ("tcl", "xdc")
        constraint_files = self._get_file_list(
            folders=folders,
            file_endings=file_endings,
            files_include=files_include,
            files_avoid=files_avoid,
        )

        constraints = []
        if constraint_files:
            synthesis_files = self.get_synthesis_files()
            for constraint_file in constraint_files:
                # Scoped constraints often depend on clocks having been created by another
                # constraint file before they can work. Set processing order to "late" to make
                # this more probable.
                constraint = Constraint(
                    constraint_file, scoped_constraint=True, processing_order="late"
                )
                constraint.validate_scoped_entity(synthesis_files)
                constraints.append(constraint)

        return constraints

    def setup_vunit(self, vunit_proj: Any, **kwargs: Any) -> None:
        """
        Setup local configuration of this module's test benches.

        .. Note::
            This default method does nothing. Should be overridden by modules that have
            any test benches that operate via generics.

        Arguments:
            vunit_proj: The VUnit project that is used to run simulation.
            kwargs: Use this to pass an arbitrary list of arguments from your ``simulate.py``
                to the module where you set up your tests. This could be, e.g., data dimensions,
                location of test files, etc.
        """

    def pre_build(
        self, project: "VivadoProject", **kwargs: Any
    ) -> bool:  # pylint: disable=unused-argument
        """
        This method hook will be called before an FPGA build is run. A typical use case for this
        mechanism is to set a register constant or default value based on the generics that
        are passed to the project. Could also be used to, e.g., generate BRAM init files
        based on project information, etc.

        .. Note::
            This default method does nothing. Should be overridden by modules that
            utilize this mechanism.

        Arguments:
            project: The project that is being built.
            kwargs: All other parameters to the build flow. Includes arguments to
                :meth:`.VivadoProject.build` method as well as other arguments set in
                :meth:`.VivadoProject.__init__`.

        Return:
            True if everything went well.
        """
        return True

    def get_build_projects(self) -> list["VivadoProject"]:
        """
        Get FPGA build projects defined by this module.

        .. Note::
            This default method does nothing. Should be overridden by modules that set up
            build projects.

        Return:
            FPGA build projects.
        """
        return []

    @staticmethod
    def test_case_name(
        name: Optional[str] = None, generics: Optional[dict[str, Any]] = None
    ) -> str:
        """
        Construct a string suitable for naming test cases.

        Arguments:
            name: Optional base name.
            generics: Dictionary of values that will be included in the name.

        Return:
            For example ``MyBaseName.GenericA_ValueA.GenericB_ValueB``.
        """
        if name:
            test_case_name = name
        else:
            test_case_name = ""

        if generics:
            generics_string = ".".join([f"{key}_{value}" for key, value in generics.items()])

            if test_case_name:
                test_case_name = f"{name}.{generics_string}"
            else:
                test_case_name = generics_string

        return test_case_name

    def add_vunit_config(  # pylint: disable=too-many-arguments
        self,
        test: Any,
        name: Optional[str] = None,
        generics: Optional[dict[str, Any]] = None,
        set_random_seed: Optional[Union[bool, int]] = False,
        pre_config: Optional[Callable[..., bool]] = None,
        post_check: Optional[Callable[..., bool]] = None,
    ) -> None:
        """
        Add config for VUnit test case.
        Wrapper that sets a suitable name and can set a random seed generic.

        Arguments:
            test: VUnit test object. Can be testbench or test case.
            name: Optional designated name for this config. Will be used to form the name of
                the config together with the ``generics`` value.
            generics: Generic values that will be applied to the testbench entity. The values
                will also be used to form the name of the config.
            set_random_seed: Controls setting of the ``seed`` generic:

                * When this argument is not assigned, or assigned ``False``, the generic will not
                  be set.
                * When set to boolean ``True``, a random natural (non-negative integer)
                  generic value will be set.
                * When set to an integer value, that value will be set for the generic.
                  This is useful to get a static test case name for waveform inspection.

                If the generic is to be set it must exist in the testbench entity, and should have
                VHDL type ``natural``.
            pre_config: Function to be run before the test.
                See `VUnit documentation <https://vunit.github.io/py/ui.html>`_ for details.
            post_check: Function to be run after the test.
                See `VUnit documentation <https://vunit.github.io/py/ui.html>`_ for details.
        """
        generics = {} if generics is None else generics

        # Note that "bool" is a sub-class of "int" in python, so isinstance(set_random_seed, int)
        # returns True if it is an integer or a bool.
        if isinstance(set_random_seed, bool):
            if set_random_seed:
                # Use the maximum range for a natural in VHDL-2008
                generics["seed"] = random.randint(0, 2**31 - 1)

        elif isinstance(set_random_seed, int):
            generics["seed"] = set_random_seed

        name = self.test_case_name(name=name, generics=generics)
        # VUnit does not allow an empty name, which can happen if both 'name' and 'generics' to
        # this method are None, but the user sets for example a 'pre_config'.
        # Avoid this error mode by setting a default name when it happens.
        name = "test" if name == "" else name

        test.add_config(name=name, generics=generics, pre_config=pre_config, post_check=post_check)

    def __str__(self) -> str:
        return f"{self.name}:{self.path}"


def get_modules(
    modules_folders: list[Path],
    names_include: Optional[set[str]] = None,
    names_avoid: Optional[set[str]] = None,
    library_name_has_lib_suffix: bool = False,
    default_registers: Optional[list[Register]] = None,
) -> ModuleList:
    """
    Get a list of Module objects based on the source code folders.

    Arguments:
        modules_folders: A list of paths where your modules are located.
        names_include: If specified, only modules with these names will be included.
        names_avoid: If specified, modules with these names will be discarded.
        library_name_has_lib_suffix: If set, the library name will be ``<module name>_lib``,
            otherwise it is just ``<module name>``.
        default_registers: Default registers.

    Return:
        List of module objects (:class:`BaseModule` or subclasses thereof)
        created from the specified folders.
    """
    modules = ModuleList()

    for module_folder in _iterate_module_folders(modules_folders):
        module_name = module_folder.name

        if names_include is not None and module_name not in names_include:
            continue

        if names_avoid is not None and module_name in names_avoid:
            continue

        modules.append(
            _get_module_object(
                path=module_folder,
                name=module_name,
                library_name_has_lib_suffix=library_name_has_lib_suffix,
                default_registers=default_registers,
            )
        )

    return modules


def _iterate_module_folders(modules_folders: list[Path]) -> Iterable[Path]:
    for modules_folder in modules_folders:
        for module_folder in modules_folder.glob("*"):
            if module_folder.is_dir():
                yield module_folder


def _get_module_object(
    path: Path,
    name: str,
    library_name_has_lib_suffix: bool,
    default_registers: Optional[list["Register"]],
) -> BaseModule:
    module_file = path / f"module_{name}.py"
    library_name = f"{name}_lib" if library_name_has_lib_suffix else name

    if module_file.exists():
        # We assume that the user lets their 'Module' class inherit from 'BaseModule'.
        module: BaseModule = load_python_module(module_file).Module(
            path=path,
            library_name=library_name,
            default_registers=default_registers,
        )
        return module

    return BaseModule(path=path, library_name=library_name, default_registers=default_registers)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# First party libraries
from tsfpga.module import BaseModule


class Module(BaseModule):
    version = 3

    def registers_hook(self):
        # Should have some registers already from the TOML file.
        register_list = self.registers
        assert register_list is not None

        register_list.add_constant(
            "version", self.version, f"Version number for the {self.name} module."
        )
        register_list.get_register("version").get_field("version").default_value = self.version



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

# Third party libraries
from git.repo import Repo

if TYPE_CHECKING:
    # Local folder libraries
    from .module_list import ModuleList


class GitSimulationSubset:
    """
    Find a subset of testbenches to simulate based on git history.
    """

    _re_tb_filename = re.compile(r"(tb_.+\.vhd)|(.+\_tb.vhd)")

    def __init__(
        self,
        repo_root: Path,
        reference_branch: str,
        vunit_proj: Any,
        vunit_preprocessed_path: Optional[Path] = None,
        modules: Optional["ModuleList"] = None,
    ) -> None:
        """
        Arguments:
            repo_root: Root directory where git commands will be run.
            reference_branch: What git branch to compare against, when finding what files have
                changed. Typically "origin/main" or "origin/master".
            vunit_proj: A vunit project with all source files and testbenches added. Will be used
                for dependency scanning.
            vunit_preprocessed_path: If location/check preprocessing is enabled
                in your VUnit project, supply the path to vunit_out/preprocessed.
            modules: A list of modules that are included in the VUnit project.
                Must be supplied only if preprocessing is enabled.
        """
        self._repo_root = repo_root
        self._reference_branch = reference_branch
        self._vunit_proj = vunit_proj
        self._vunit_preprocessed_path = vunit_preprocessed_path
        self._modules = modules

        if (vunit_preprocessed_path is not None) != (modules is not None):
            raise ValueError("Can not supply only one of vunit_preprocessed_path and modules")

    def find_subset(self) -> list[tuple[str, str]]:
        """
        Return all testbenches that have changes, or depend on files that have changes.

        Return:
            The testbench names and their corresponding library names.
            A list of tuples ("testbench name", "library name").
        """
        diff_files = self._find_diff_vhd_files()

        if self._vunit_preprocessed_path:
            # If preprocessing is enabled, VUnit's dependency graph is based on the files that
            # are in the vunit_out/preprocessed folder, not in the file's original location.
            # So manipulate the paths to point there.
            diff_files = self._get_preprocessed_file_locations(diff_files)
            self._print_file_list("Resolved diff file locations to be", diff_files)

        # Find all testbench files that are available
        testbenches = self._find_testbenches()

        # Gather the testbenches that depend on any files that have diffs
        testbenches_to_run = []
        for testbench_source_file, library_name in testbenches:
            if self._source_file_depends_on_files(
                source_file=testbench_source_file,
                files=diff_files,
            ):
                testbench_file_name = Path(testbench_source_file.name).stem
                testbenches_to_run.append((testbench_file_name, library_name))

        return testbenches_to_run

    def _find_diff_vhd_files(self) -> set[Path]:
        repo = Repo(self._repo_root)

        head_commit = repo.head.commit
        reference_commit = repo.commit(self._reference_branch)

        # Local uncommitted changed
        working_tree_changes = head_commit.diff(None)

        # Changes in the git log compared to the reference commit
        history_changes = head_commit.diff(reference_commit)

        return self._iterate_vhd_file_diffs(diffs=working_tree_changes + history_changes)

    def _iterate_vhd_file_diffs(self, diffs: Any) -> set[Path]:
        """
        Return the currently existing files that have been changed (added/renamed/modified)
        within any of the diffs commits.
        """
        files = set()

        for diff in diffs:
            # The diff contains "a" -> "b" changes information. In case of file deletion, a_path
            # will be set but not b_path. Removed files are not included by this method.
            if diff.b_path is not None:
                b_path = Path(diff.b_path)

                # A file can be changed in an early commit, but then removed/renamed in a
                # later commit. Include only files that are currently existing.
                if b_path.exists():
                    if b_path.name.endswith(".vhd"):
                        files.add(b_path.resolve())

        self._print_file_list("Found git diff in the following files", files)
        return files

    def _get_preprocessed_file_locations(self, vhd_files: set[Path]) -> set[Path]:
        """
        Find the location of a VUnit preprocessed file, based on the path in the modules tree.
        Not all VHDL files are included in the simulation projects (e.g. often files that depend
        on IP cores are excluded), hence files that can not be found in any module's simulation
        files are ignored.
        """
        result = set()
        for vhd_file in vhd_files:
            library_name = self._get_library_name_from_path(vhd_file)

            if library_name is not None:
                # Ignore that '_vunit_preprocessed_path' is type 'Path | None', since we only come
                # if it is not 'None'.
                preprocessed_file = (
                    self._vunit_preprocessed_path  # type: ignore[operator]
                    / library_name
                    / vhd_file.name
                )
                assert preprocessed_file.exists(), preprocessed_file

                result.add(preprocessed_file)

        return result

    def _get_library_name_from_path(self, vhd_file: Path) -> Optional[str]:
        """
        Returns (str): Library name for the given file path.
            Will return None if no library can be found.
        """
        # Ignore that '_modules' is type 'Path | None', since we only come
        # if it is not 'None'.
        for module in self._modules:  # type: ignore[union-attr]
            for module_hdl_file in module.get_simulation_files(include_ip_cores=True):
                if module_hdl_file.path.name == vhd_file.name:
                    return module.library_name

        print(f"Could not find library for file {vhd_file}. It will be skipped.")
        return None

    def _find_testbenches(self) -> list[tuple[Any, str]]:
        """
        Find all testbench files that are available in the VUnit project.

        Return:
            The VUnit ``SourceFile`` objects and library names for the files.
        """
        result = []
        for source_file in self._vunit_proj.get_source_files():
            source_file_path = Path(source_file.name)
            assert source_file_path.exists(), source_file_path

            # The file is considered a testbench if it follows the tb naming pattern
            if re.fullmatch(self._re_tb_filename, source_file_path.name):
                result.append((source_file, source_file.library.name))

        return result

    def _source_file_depends_on_files(self, source_file: Any, files: set[Path]) -> bool:
        """
        Return True if the source_file depends on any of the files.
        """
        # Note that this includes the source_file itself. Is a list of SourceFile objects.
        implementation_subset = self._vunit_proj.get_implementation_subset([source_file])

        # Convert to a set of absolute Paths, for comparison with "files" which is of that type.
        source_file_dependencies = {
            Path(implementation_file.name).resolve()
            for implementation_file in implementation_subset
        }

        intersection = source_file_dependencies & files
        if not intersection:
            return False

        self._print_file_list(
            f"Testbench {source_file.name} depends on the following files which have a diff",
            intersection,
        )
        return True

    @staticmethod
    def _print_file_list(title: str, files: set[Path]) -> None:
        print(f"{title}:")
        for file_path in files:
            print(f"  {file_path}")
        print()



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import re
from pathlib import Path
from typing import Any, Optional

# First party libraries
from tsfpga import DEFAULT_FILE_ENCODING
from tsfpga.system_utils import create_directory, run_command, system_is_windows

# Local folder libraries
from .simlib_common import VivadoSimlibCommon


class VivadoSimlibGhdl(VivadoSimlibCommon):
    """
    Handle Vivado simlib with GHDL.
    """

    library_names = ["unisim", "secureip", "unimacro", "unifast"]

    def __init__(
        self,
        vivado_path: Optional[Path],
        output_path: Path,
        vunit_proj: Any,
        simulator_interface: Any,
    ) -> None:
        """
        Arguments:
            output_path: The compiled simlib will be placed here.
            vunit_proj: The VUnit project that is used to run simulation.
            simulator_interface: A VUnit SimulatorInterface class.
            vivado_path: Path to Vivado executable.
        """
        self.ghdl_binary = Path(simulator_interface.find_prefix()) / "ghdl"

        super().__init__(vivado_path=vivado_path, output_path=output_path)

        self._vunit_proj = vunit_proj

    def _compile(self) -> None:
        self._compile_unisim()
        self._compile_secureip()
        self._compile_unimacro()
        self._compile_unifast()

    def _compile_unisim(self) -> None:
        library_path = self._libraries_path / "unisims"

        vhd_files = []

        for vhd_file_base in [
            "unisim_VPKG",
            "unisim_retarget_VCOMP",
        ]:
            vhd_file = library_path / f"{vhd_file_base}.vhd"
            assert vhd_file.exists(), vhd_file

            vhd_files.append(vhd_file)

        primitive_dir = library_path / "primitive"
        vhd_files += self._get_compile_order(library_path=primitive_dir)

        retarget_dir = library_path / "retarget"
        for vhd_file in retarget_dir.glob("*.vhd"):
            vhd_files.append(vhd_file)

        self._compile_ghdl(vhd_files=vhd_files, library_name="unisim")

    def _compile_secureip(self) -> None:
        library_path = self._libraries_path / "unisims" / "secureip"

        vhd_files = []
        for vhd_file in library_path.glob("*.vhd"):
            vhd_files.append(vhd_file)

        self._compile_ghdl(vhd_files=vhd_files, library_name="secureip")

    def _compile_unimacro(self) -> None:
        library_path = self._libraries_path / "unimacro"

        vhd_files = []

        vhd_file = library_path / "unimacro_VCOMP.vhd"
        assert vhd_file.exists(), vhd_file
        vhd_files.append(vhd_file)

        vhd_files += self._get_compile_order(library_path=library_path)

        self._compile_ghdl(vhd_files=vhd_files, library_name="unimacro")

    def _compile_unifast(self) -> None:
        library_path = self._libraries_path / "unifast" / "primitive"
        vhd_files = self._get_compile_order(library_path=library_path)

        self._compile_ghdl(vhd_files=vhd_files, library_name="unifast")

    @staticmethod
    def _get_compile_order(library_path: Path) -> list[Path]:
        """
        Get compile order (list of file paths, in order) from an existing compile order
        file provided by Xilinx.
        """
        vhd_files = []

        with open(
            library_path / "vhdl_analyze_order", encoding=DEFAULT_FILE_ENCODING
        ) as file_handle:
            for vhd_file_base in file_handle.readlines():
                vhd_file = library_path / vhd_file_base.strip()
                assert vhd_file.exists(), vhd_file
                vhd_files.append(vhd_file)

        return vhd_files

    def _compile_ghdl(self, vhd_files: list[Path], library_name: str) -> None:
        """
        Compile a list of files into the specified library.
        """
        workdir = self.output_path / library_name
        create_directory(workdir, empty=False)

        vhd_paths_str = [str(vhd_file) for vhd_file in vhd_files]
        # We print a list of the files that will be compiled.
        # Use relative paths to the Vivado path, in order to keep it a little shorter
        # (it is still massively long).
        vhd_paths_relative = [
            str(vhd_file.relative_to(self._libraries_path)) for vhd_file in vhd_files
        ]

        def print_compiling(path: str) -> None:
            print(f"Compiling {path} into {library_name}...")

        def execute_ghdl(files: list[str]) -> None:
            self._execute_ghdl(workdir=workdir, library_name=library_name, files=files)

        # There seems to be a command length limit on Windows.
        # While compiling all files in one command gives a huge performance boost
        # (on Linux with GCC backend at least, as far as we know) the resulting command is in
        # the order of 90k characters long.
        # This does not seem to work on Windows.
        # So we auto detect the OS to work around this limitation, while keeping the performance
        # boost on Linux.
        # See https://gitlab.com/tsfpga/tsfpga/-/merge_requests/499
        compile_file_by_file = system_is_windows()

        if compile_file_by_file:
            # Compile each file in a separate command.
            for vhd_file_idx, vhd_file_str in enumerate(vhd_paths_str):
                print_compiling(vhd_paths_relative[vhd_file_idx])
                execute_ghdl(files=[vhd_file_str])

        else:
            # Compile all files in one single command.
            paths_to_print = ", ".join(vhd_paths_relative)
            print_compiling(paths_to_print)
            execute_ghdl(files=vhd_paths_str)

    def _execute_ghdl(self, workdir: Path, library_name: str, files: list[str]) -> None:
        cmd = [
            str(self.ghdl_binary),
            "-a",
            "--ieee=synopsys",
            "--std=08",
            f"--workdir={str(workdir.resolve())}",
            f"-P{str(self.output_path / 'unisim')}",
            "-fexplicit",
            "-frelaxed-rules",
            "--no-vital-checks",
            "--warn-binding",
            "--mb-comments",
            f"--work={library_name}",
        ] + files

        run_command(cmd, cwd=self.output_path)

    def _get_simulator_tag(self) -> str:
        """
        Return simulator version tag as a string.
        """
        cmd = [str(self.ghdl_binary), "--version"]
        output = run_command(cmd, capture_output=True).stdout

        regexp_with_tag = re.compile(r"^GHDL (\S+) \((\S+)\).*")
        match = regexp_with_tag.search(output)
        if match is not None:
            return self._format_version(f"ghdl_{match.group(1)}_{match.group(2)}")

        regexp_without_tag = re.compile(r"^GHDL (\S+).*")
        match = regexp_without_tag.search(output)
        if match is not None:
            return self._format_version(f"ghdl_{match.group(1)}")

        raise ValueError(f"Could not find GHDL version string: {output}")

    def _add_to_vunit_project(self) -> None:
        """
        Add the compiled simlib to your VUnit project.
        """
        for library_name in self.library_names:
            library_path = self.output_path / library_name
            assert library_path.exists(), library_path

            self._vunit_proj.add_external_library(library_name, library_path)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from collections import OrderedDict
from pathlib import Path
from unittest.mock import MagicMock

# Third party libraries
import pytest

# First party libraries
from tsfpga.build_step_tcl_hook import BuildStepTclHook
from tsfpga.ip_core_file import IpCoreFile
from tsfpga.module import BaseModule, get_modules
from tsfpga.system_utils import create_file

# pylint: disable=unused-import
from tsfpga.test.test_utils import file_contains_string
from tsfpga.vivado.common import to_tcl_path
from tsfpga.vivado.generics import BitVectorGenericValue, StringGenericValue
from tsfpga.vivado.tcl import VivadoTcl


def test_set_create_run_index():
    tcl = VivadoTcl(name="").create(project_folder=Path(), modules=[], part="", top="", run_index=2)
    assert "\ncurrent_run [get_runs synth_2]\n" in tcl


def test_static_generics():
    # Use OrderedDict here in test so that order will be preserved and we can test for equality.
    # In real world case a normal dict can be used.
    generics = OrderedDict(
        enable=True,
        disable=False,
        integer=123,
        slv=BitVectorGenericValue("0101"),
        string=StringGenericValue("apa"),
    )

    tcl = VivadoTcl(name="").create(
        project_folder=Path(), modules=[], part="", top="", run_index=1, generics=generics
    )
    expected = (
        "\nset_property generic {enable=1'b1 disable=1'b0 integer=123 slv=4'b0101 string=\"apa\"} "
        "[current_fileset]\n"
    )
    assert expected in tcl


def test_build_step_hooks():
    dummy = BuildStepTclHook(Path("dummy.tcl"), "STEPS.SYNTH_DESIGN.TCL.PRE")
    files = BuildStepTclHook(Path("files.tcl"), "STEPS.ROUTE_DESIGN.TCL.PRE")
    tcl = VivadoTcl(name="").create(
        project_folder=Path(),
        modules=[],
        part="part",
        top="",
        run_index=1,
        build_step_hooks=[dummy, files],
    )

    assert (
        f"\nset_property STEPS.SYNTH_DESIGN.TCL.PRE {{{to_tcl_path(dummy.tcl_file)}}} ${{run}}\n"
        in tcl
    )
    assert (
        f"\nset_property STEPS.ROUTE_DESIGN.TCL.PRE {{{to_tcl_path(files.tcl_file)}}} ${{run}}\n"
        in tcl
    )


def test_build_step_hooks_with_same_hook_step(tmp_path):
    dummy = BuildStepTclHook(Path("dummy.tcl"), "STEPS.SYNTH_DESIGN.TCL.PRE")
    files = BuildStepTclHook(Path("files.tcl"), "STEPS.SYNTH_DESIGN.TCL.PRE")
    tcl = VivadoTcl(name="").create(
        project_folder=tmp_path / "dummy_project_folder",
        modules=[],
        part="part",
        top="",
        run_index=1,
        build_step_hooks=[dummy, files],
    )

    hook_file = tmp_path / "dummy_project_folder" / "hook_STEPS_SYNTH_DESIGN_TCL_PRE.tcl"

    assert file_contains_string(hook_file, f"source {{{to_tcl_path(dummy.tcl_file)}}}")
    assert file_contains_string(hook_file, f"source {{{to_tcl_path(files.tcl_file)}}}")

    assert (
        f"\nset_property STEPS.SYNTH_DESIGN.TCL.PRE {{{to_tcl_path(hook_file)}}} ${{run}}\n" in tcl
    )


def test_ip_cache_location(tmp_path):
    tcl = VivadoTcl(name="").create(
        project_folder=Path(), modules=[], part="part", top="", run_index=1
    )
    assert "config_ip_cache" not in tcl

    tcl = VivadoTcl(name="").create(
        project_folder=Path(), modules=[], part="part", top="", run_index=1, ip_cache_path=tmp_path
    )
    assert f"\nconfig_ip_cache -use_cache_location {{{to_tcl_path(tmp_path)}}}\n" in tcl


def test_multiple_threads_is_capped_by_vivado_limits():
    num_threads = 128
    tcl = VivadoTcl(name="").build(
        project_file=Path(), output_path=Path(), num_threads=num_threads, run_index=1
    )
    print(tcl)
    assert "set_param general.maxThreads 32" in tcl
    assert "set_param synth.maxThreads 8" in tcl
    assert "launch_runs ${run} -jobs 128" in tcl
    assert "launch_runs ${run} -jobs 128" in tcl


def test_set_build_run_index():
    tcl = VivadoTcl(name="").build(
        project_file=Path(), output_path=Path(), num_threads=0, run_index=1
    )
    assert "impl_1" in tcl
    assert "synth_1" in tcl
    assert "impl_2" not in tcl
    assert "synth_2" not in tcl

    tcl = VivadoTcl(name="").build(
        project_file=Path(), output_path=Path(), num_threads=0, run_index=2
    )
    assert "impl_2" in tcl
    assert "synth_2" in tcl
    assert "impl_1" not in tcl
    assert "synth_1" not in tcl


def test_runtime_generics():
    tcl = VivadoTcl(name="").build(
        project_file=Path(),
        output_path=Path(),
        num_threads=0,
        run_index=0,
        generics=dict(dummy=True),
    )
    expected = "\nset_property generic {dummy=1'b1} [current_fileset]\n"
    assert expected in tcl


def test_build_with_synth_only():
    tcl = VivadoTcl(name="").build(
        project_file=Path(), output_path=Path(), num_threads=0, run_index=0, synth_only=False
    )
    assert "synth_" in tcl
    assert "impl_" in tcl

    tcl = VivadoTcl(name="").build(
        project_file=Path(), output_path=Path(), num_threads=0, run_index=0, synth_only=True
    )
    assert "synth_" in tcl
    assert "impl_" not in tcl


def test_build_with_from_impl():
    tcl = VivadoTcl(name="").build(
        project_file=Path(), output_path=Path(), num_threads=0, run_index=0, from_impl=False
    )
    assert "synth_" in tcl
    assert "impl_" in tcl

    tcl = VivadoTcl(name="").build(
        project_file=Path(), output_path=Path(), num_threads=0, run_index=0, from_impl=True
    )
    assert "synth_" not in tcl
    assert "impl_" in tcl


def test_module_getters_are_called_with_correct_arguments():
    modules = [MagicMock(spec=BaseModule)]
    VivadoTcl(name="").create(
        project_folder=Path(),
        modules=modules,
        part="",
        top="",
        run_index=1,
        other_arguments=dict(apa=123, hest=456),
    )

    modules[0].get_synthesis_files.assert_called_once_with(apa=123, hest=456)
    modules[0].get_scoped_constraints.assert_called_once_with(apa=123, hest=456)
    modules[0].get_ip_core_files.assert_called_once_with(apa=123, hest=456)


@pytest.fixture
def vivado_tcl_test(tmp_path):
    class VivadoTclTest:  # pylint: disable=too-many-instance-attributes
        def __init__(self):
            self.modules_folder = tmp_path / "modules"

            # A library with some synth files and some test files
            self.a_vhd = to_tcl_path(create_file(self.modules_folder / "apa" / "a.vhd"))
            self.b_vhd = to_tcl_path(create_file(self.modules_folder / "apa" / "b.vhd"))
            self.tb_a_vhd = to_tcl_path(
                create_file(self.modules_folder / "apa" / "test" / "tb_a.vhd")
            )
            self.a_xdc = to_tcl_path(
                create_file(self.modules_folder / "apa" / "scoped_constraints" / "a.xdc")
            )

            self.c_v = to_tcl_path(create_file(self.modules_folder / "apa" / "c.v"))
            self.b_tcl = to_tcl_path(
                create_file(self.modules_folder / "apa" / "scoped_constraints" / "b.tcl")
            )

            self.c_tcl = to_tcl_path(
                create_file(self.modules_folder / "apa" / "ip_cores" / "c.tcl")
            )

            # A library with only test files
            self.d_vhd = to_tcl_path(create_file(self.modules_folder / "zebra" / "test" / "d.vhd"))

            self.modules = get_modules([self.modules_folder])

            self.tcl = VivadoTcl(name="name")

    return VivadoTclTest()


# False positive for pytest fixtures
# pylint: disable=redefined-outer-name


def test_source_file_list_is_correctly_formatted(vivado_tcl_test):
    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(), modules=vivado_tcl_test.modules, part="", top="", run_index=1
    )

    # Order of files is not really deterministic
    expected_1 = (
        "\nread_vhdl -library apa -vhdl2008 "
        f"{{{{{vivado_tcl_test.b_vhd}}} {{{vivado_tcl_test.a_vhd}}}}}\n"
    )
    expected_2 = (
        "\nread_vhdl -library apa -vhdl2008 "
        f"{{{{{vivado_tcl_test.a_vhd}}} {{{vivado_tcl_test.b_vhd}}}}}\n"
    )
    assert expected_1 in tcl or expected_2 in tcl

    expected = f"\nread_verilog {{{vivado_tcl_test.c_v}}}\n"
    assert expected in tcl


def test_only_synthesis_files_added_to_create_project_tcl(vivado_tcl_test):
    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(), modules=vivado_tcl_test.modules, part="", top="", run_index=1
    )
    assert vivado_tcl_test.a_vhd in tcl and vivado_tcl_test.c_v in tcl
    assert vivado_tcl_test.tb_a_vhd not in tcl and "tb_a.vhd" not in tcl


def test_constraints(vivado_tcl_test):
    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(), modules=vivado_tcl_test.modules, part="part", top="", run_index=1
    )

    expected = f"\nread_xdc -ref a {{{vivado_tcl_test.a_xdc}}}\n"
    assert expected in tcl
    expected = f"\nread_xdc -ref b -unmanaged {{{vivado_tcl_test.b_tcl}}}\n"
    assert expected in tcl


def test_ip_core_files(vivado_tcl_test):
    ip_core_file_path = vivado_tcl_test.modules_folder.parent / "my_name.tcl"
    module = MagicMock(spec=BaseModule)
    module.get_ip_core_files.return_value = [
        IpCoreFile(path=ip_core_file_path, apa="hest", zebra=123)
    ]

    vivado_tcl_test.modules.append(module)

    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(), modules=vivado_tcl_test.modules, part="part", top="", run_index=1
    )

    assert (
        f"""
proc create_ip_core_c {{}} {{
  source -notrace {{{vivado_tcl_test.c_tcl}}}
}}
create_ip_core_c
"""
        in tcl
    )

    assert (
        f"""
proc create_ip_core_my_name {{}} {{
  set apa "hest"
  set zebra "123"
  source -notrace {{{to_tcl_path(ip_core_file_path)}}}
}}
create_ip_core_my_name
"""
        in tcl
    )


def test_create_with_ip_cores_only(vivado_tcl_test):
    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(),
        modules=vivado_tcl_test.modules,
        part="part",
        top="",
        run_index=1,
        ip_cores_only=True,
    )
    assert vivado_tcl_test.c_tcl in tcl
    assert vivado_tcl_test.a_vhd not in tcl


def test_empty_library_not_in_create_project_tcl(vivado_tcl_test):
    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(), modules=vivado_tcl_test.modules, part="part", top="", run_index=1
    )
    assert "zebra" not in tcl


def test_multiple_tcl_sources(vivado_tcl_test):
    extra_tcl_sources = [Path("dummy.tcl"), Path("files.tcl")]
    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(),
        modules=vivado_tcl_test.modules,
        part="part",
        top="",
        run_index=1,
        tcl_sources=extra_tcl_sources,
    )

    for filename in extra_tcl_sources:
        assert f"\nsource -notrace {{{to_tcl_path(filename)}}}\n" in tcl


def test_io_buffer_setting(vivado_tcl_test):
    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(),
        modules=vivado_tcl_test.modules,
        part="part",
        top="",
        run_index=1,
        disable_io_buffers=True,
    )

    no_io_buffers_tcl = (
        "\nset_property -name {STEPS.SYNTH_DESIGN.ARGS.MORE OPTIONS} "
        "-value -no_iobuf -objects [get_runs synth_1]\n"
    )
    assert no_io_buffers_tcl in tcl

    tcl = vivado_tcl_test.tcl.create(
        project_folder=Path(),
        modules=vivado_tcl_test.modules,
        part="part",
        top="",
        run_index=1,
        disable_io_buffers=False,
    )

    assert no_io_buffers_tcl not in tcl


def test_analyze_synthesis_settings_on_and_off(vivado_tcl_test):
    tcl = vivado_tcl_test.tcl.build(
        project_file=Path(),
        output_path=Path(),
        num_threads=1,
        run_index=1,
        analyze_synthesis_timing=True,
    )
    assert "open_run" in tcl
    assert "report_clock_interaction" in tcl

    tcl = vivado_tcl_test.tcl.build(
        project_file=Path(),
        output_path=Path(),
        num_threads=1,
        run_index=1,
        analyze_synthesis_timing=False,
    )
    # When disabled, the run should not even be opened, which saves time
    assert "open_run" not in tcl
    assert "report_clock_interaction" not in tcl


def test_impl_explore(vivado_tcl_test):
    num_runs = 4

    tcl = vivado_tcl_test.tcl.build(
        project_file=Path(),
        output_path=Path(),
        num_threads=num_runs,
        run_index=1,
        impl_explore=True,
    )

    assert f"launch_runs -jobs {num_runs} [get_runs impl_explore_*] -to_step write_bitstream" in tcl
    assert "wait_on_runs -quiet -exit_condition ANY_ONE_MET_TIMING [get_runs impl_explore_*]" in tcl
    assert 'reset_runs [get_runs -filter {STATUS == "Queued..."}]' in tcl
    assert 'wait_on_runs -quiet [get_runs -filter {STATUS != "Not started"} impl_explore_*]' in tcl
    assert 'foreach run [get_runs -filter {PROGRESS == "100%"} impl_explore_*]' in tcl



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path

# Local folder libraries
from .about import get_short_slogan

THIS_DIR = Path(__file__).parent
REPO_ROOT = THIS_DIR.parent.resolve()

TSFPGA_PATH = REPO_ROOT / "tsfpga"
TSFPGA_DOC = REPO_ROOT / "doc"
TSFPGA_TCL = THIS_DIR / "vivado" / "tcl"
TSFPGA_GENERATED = REPO_ROOT / "generated"

TSFPGA_EXAMPLES = TSFPGA_PATH / "examples"
TSFPGA_EXAMPLE_MODULES = TSFPGA_EXAMPLES / "modules"

# Default encoding when opening files
DEFAULT_FILE_ENCODING = "utf-8"

__version__ = "12.3.2"
__doc__ = get_short_slogan()  # pylint: disable=redefined-builtin



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from unittest.mock import MagicMock

# Third party libraries
import pytest

# First party libraries
from tsfpga.build_project_list import BuildProjectList
from tsfpga.module import BaseModule
from tsfpga.system_utils import create_directory
from tsfpga.vivado.project import BuildResult, VivadoProject


@pytest.fixture
def build_project_list_test():
    # pylint: disable=too-many-instance-attributes
    class TestBuildProjectList:
        @staticmethod
        def _get_mocks(name, is_netlist_build):
            project = MagicMock(spec=VivadoProject)
            project.name = name
            project.__str__.return_value = f"MockProject {name}"
            project.is_netlist_build = is_netlist_build

            module = MagicMock(spec=BaseModule)
            module.name = name
            module.get_build_projects.return_value = [project]

            return module, project

        def __init__(self):
            self.module_one, self.project_one = self._get_mocks("one", False)
            self.module_two, self.project_two = self._get_mocks("two", False)

            self.module_three, self.project_three = self._get_mocks("three", True)
            self.module_four, self.project_four = self._get_mocks("four", True)

            self.modules = [self.module_one, self.module_two, self.module_three, self.module_four]

    return TestBuildProjectList()


# False positive for pytest fixtures
# pylint: disable=redefined-outer-name


def test_can_list_without_error(build_project_list_test):
    list_str = str(BuildProjectList(build_project_list_test.modules, project_filters=[]))
    assert "one" in list_str
    assert "two" in list_str


def test_project_filtering(build_project_list_test):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=[])
    assert len(project_list.projects) == 2
    assert build_project_list_test.project_one in project_list.projects
    assert build_project_list_test.project_two in project_list.projects

    project_list = BuildProjectList(
        build_project_list_test.modules, project_filters=["apa", "*ne", "three", "four"]
    )
    assert len(project_list.projects) == 1
    assert build_project_list_test.project_one in project_list.projects

    project_list = BuildProjectList(
        build_project_list_test.modules, project_filters=["one", "one", "on*"]
    )
    assert len(project_list.projects) == 1
    assert build_project_list_test.project_one in project_list.projects

    project_list = BuildProjectList(
        build_project_list_test.modules, project_filters=[], include_netlist_not_top_builds=True
    )
    assert len(project_list.projects) == 2
    assert build_project_list_test.project_three in project_list.projects
    assert build_project_list_test.project_four in project_list.projects

    project_list = BuildProjectList(
        build_project_list_test.modules,
        include_netlist_not_top_builds=True,
        project_filters=["apa", "one", "two", "thr*"],
    )
    assert len(project_list.projects) == 1
    assert build_project_list_test.project_three in project_list.projects


def test_create(build_project_list_test, tmp_path):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one", "two"])
    assert project_list.create(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        ip_cache_path=tmp_path / "ip_cache_path",
    )

    build_project_list_test.project_one.create.assert_called_once_with(
        project_path=tmp_path / "projects_path" / "one" / "project",
        ip_cache_path=tmp_path / "ip_cache_path",
    )

    build_project_list_test.project_two.create.assert_called_once_with(
        project_path=tmp_path / "projects_path" / "two" / "project",
        ip_cache_path=tmp_path / "ip_cache_path",
    )

    build_project_list_test.project_three.create.assert_not_called()
    build_project_list_test.project_four.create.assert_not_called()


def test_create_unless_exists(build_project_list_test, tmp_path):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one"])
    assert project_list.create_unless_exists(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        ip_cache_path=tmp_path / "ip_cache_path",
    )

    build_project_list_test.project_one.create.assert_called_once_with(
        project_path=tmp_path / "projects_path" / "one" / "project",
        ip_cache_path=tmp_path / "ip_cache_path",
    )

    # Create project file manually
    create_directory(tmp_path / "projects_path" / "one" / "project")
    (tmp_path / "projects_path" / "one" / "project" / "one.xpr").write_text("")

    assert project_list.create_unless_exists(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        ip_cache_path=tmp_path / "ip_cache_path",
    )

    # Still only called once after second create_unless_exists()
    build_project_list_test.project_one.create.assert_called_once()

    build_project_list_test.project_two.create.assert_not_called()
    build_project_list_test.project_three.create.assert_not_called()
    build_project_list_test.project_four.create.assert_not_called()


def test_build(build_project_list_test, tmp_path):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one"])
    assert project_list.build(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        num_threads_per_build=4,
        other_build_argument=True,
    )

    build_project_list_test.project_one.build.assert_called_once_with(
        project_path=tmp_path / "projects_path" / "one" / "project",
        output_path=tmp_path / "projects_path" / "one",
        num_threads=4,
        other_build_argument=True,
    )


def test_build_fail_should_return_false(build_project_list_test, tmp_path):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one"])
    build_project_list_test.project_one.build.return_value = MagicMock(spec=BuildResult)
    build_project_list_test.project_one.build.return_value.success = False

    assert not project_list.build(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        num_threads_per_build=4,
        other_build_argument=True,
    )


def test_build_with_output_path(build_project_list_test, tmp_path):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one"])
    assert project_list.build(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        num_threads_per_build=4,
        output_path=tmp_path / "output_path",
    )

    build_project_list_test.project_one.build.assert_called_once_with(
        project_path=tmp_path / "projects_path" / "one" / "project",
        output_path=tmp_path / "output_path" / "one",
        num_threads=4,
    )


def test_build_with_collect_artifacts(build_project_list_test, tmp_path):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one"])
    collect_artifacts = MagicMock()
    assert project_list.build(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        num_threads_per_build=4,
        collect_artifacts=collect_artifacts,
    )

    collect_artifacts.assert_called_once_with(
        project=build_project_list_test.project_one,
        output_path=tmp_path / "projects_path" / "one",
    )


def test_build_with_collect_artifacts_and_output_path(build_project_list_test, tmp_path):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one"])
    collect_artifacts = MagicMock()
    assert project_list.build(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        num_threads_per_build=4,
        output_path=tmp_path / "output_path",
        collect_artifacts=collect_artifacts,
    )

    collect_artifacts.assert_called_once_with(
        project=build_project_list_test.project_one,
        output_path=tmp_path / "output_path" / "one",
    )


def test_build_with_collect_artifacts_return_false_should_fail_build(
    build_project_list_test, tmp_path
):
    project_list = BuildProjectList(build_project_list_test.modules, project_filters=["one"])
    collect_artifacts = MagicMock()
    collect_artifacts.return_value = False
    assert not project_list.build(
        projects_path=tmp_path / "projects_path",
        num_parallel_builds=2,
        num_threads_per_build=4,
        collect_artifacts=collect_artifacts,
    )


def test_open(build_project_list_test, tmp_path):
    project_list = BuildProjectList(
        build_project_list_test.modules, project_filters=[], include_netlist_not_top_builds=True
    )
    assert project_list.open(projects_path=tmp_path / "projects_path")

    build_project_list_test.project_three.open.assert_called_once_with(
        project_path=tmp_path / "projects_path" / "three" / "project"
    )
    build_project_list_test.project_four.open.assert_called_once_with(
        project_path=tmp_path / "projects_path" / "four" / "project"
    )
    build_project_list_test.project_one.open.assert_not_called()
    build_project_list_test.project_two.open.assert_not_called()
